from enum import IntEnum

from sonolus.script.containers import VarArray
from tests.script.conftest import run_and_validate


class Color(IntEnum):
    RED = 1
    GREEN = 2
    BLUE = 3


def test_len_int_enum():
    def fn():
        return len(Color)

    assert run_and_validate(fn) == len(Color)


def test_iterate_int_enum():
    n = len(Color)

    def fn():
        results = VarArray[int, n].new()
        for c in Color:
            results.append(c)
        return results

    assert list(run_and_validate(fn)) == [c.value for c in Color]



from sonolus.script.instruction import (
    StandardInstruction,
    StandardInstructionIcon,
    instruction_icons,
    instructions,
)


@instructions
class Instructions:
    tap: StandardInstruction.TAP
    tap_flick: StandardInstruction.TAP_FLICK
    tap_hold: StandardInstruction.TAP_HOLD
    hold_follow: StandardInstruction.HOLD_FOLLOW
    release: StandardInstruction.RELEASE
    hold_flick: StandardInstruction.FLICK


@instruction_icons
class InstructionIcons:
    hand: StandardInstructionIcon.HAND



from sonolus.backend.blocks import BlockData
from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet
from sonolus.backend.ops import Op
from sonolus.backend.optimize.dominance import DominanceFrontiers
from sonolus.backend.optimize.flow import BasicBlock, traverse_cfg_preorder
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig
from sonolus.backend.place import BlockPlace, SSAPlace

# Limited in what we can do here since floating-point math is not commutative
COMMUTATIVE_OPS = frozenset({Op.Equal, Op.NotEqual, Op.Max, Op.Min})


def _sort_key(expr):
    match expr:
        case IRConst(value=value):
            return (0, value)
        case IRGet(place=SSAPlace(name=name, num=num)):
            return (1, 0, name, num)
        case IRGet(place=BlockPlace(block=block, index=index, offset=offset)):
            index_key = (
                _sort_key(index)
                if isinstance(index, IRConst | IRPureInstr | IRGet)
                else ((index.name, index.num) if isinstance(index, SSAPlace) else (index,))
            )
            return (1, 1, int(block) if isinstance(block, int) else 0, index_key, offset)
        case IRPureInstr(op=op, args=args):
            return (2, op.value, tuple(_sort_key(arg) for arg in args))
        case _:
            return (3,)


def _cost(expr) -> int:
    match expr:
        case IRConst():
            return 1
        case IRGet(place=SSAPlace()):
            return 3
        case IRGet(place=BlockPlace(index=index)):
            return 2 + _cost(index)
        case IRPureInstr(args=args):
            return 1 + sum(_cost(arg) for arg in args)
        case int():
            return 1
        case SSAPlace():
            return 3
        case _:
            return 0


class CommonSubexpressionElimination(CompilerPass):
    """Global CSE using the dominator tree with subexpression extraction.

    Walks the dominator tree in preorder. Every CSE-candidate pure expression
    with cost > 4 is extracted into an SSA variable on first sight. Duplicates
    are replaced with a reference to the existing variable. A later InlineVars
    pass cleans up any single-use extractions.

    For top-level IRSet(SSAPlace, expr), the existing variable is reused (free).

    Also canonicalizes commutative operations by sorting their arguments.
    """

    def requires(self) -> set[CompilerPass]:
        return {DominanceFrontiers()}

    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        callback = config.callback

        # Phase 1: Canonicalize all expressions
        for block in traverse_cfg_preorder(entry):
            block.statements = [self._canonicalize_stmt(stmt, callback) for stmt in block.statements]
            block.test = self._canonicalize_stmt(block.test, callback)

        # Phase 2: Rewrite with extraction
        next_id = [0]
        self._process_block(entry, {}, callback, next_id)
        return entry

    def _is_cse_candidate(self, expr, callback: str | None) -> bool:
        match expr:
            case IRConst():
                return True
            case IRGet(place=SSAPlace()):
                return True
            case IRGet(place=BlockPlace(block=block, index=index)):
                return (
                    isinstance(block, BlockData)
                    and callback not in block.writable
                    and (isinstance(index, int | SSAPlace) or self._is_cse_candidate(index, callback))
                )
            case IRPureInstr(op=op, args=args):
                return op.pure and not op.side_effects and all(self._is_cse_candidate(arg, callback) for arg in args)
            case _:
                return False

    def _canonicalize(self, expr, callback: str | None):
        if not self._is_cse_candidate(expr, callback):
            return expr
        match expr:
            case IRPureInstr(op=op, args=args):
                args = [self._canonicalize(arg, callback) for arg in args]
                if op in COMMUTATIVE_OPS:
                    args = sorted(args, key=_sort_key)
                return IRPureInstr(op=op, args=args)
            case _:
                return expr

    def _canonicalize_stmt(self, stmt, callback: str | None):
        match stmt:
            case IRPureInstr():
                return self._canonicalize(stmt, callback)
            case IRInstr(op=op, args=args):
                return IRInstr(op=op, args=[self._canonicalize_stmt(arg, callback) for arg in args])
            case IRSet(place=place, value=value):
                return IRSet(place=place, value=self._canonicalize_stmt(value, callback))
            case _:
                return stmt

    def _process_block(
        self,
        block: BasicBlock,
        available: dict[IRPureInstr, SSAPlace],
        callback: str | None,
        next_id: list[int],
    ):
        added: list[IRPureInstr] = []

        new_statements: list = []
        for stmt in block.statements:
            pre_stmts: list[IRSet] = []
            stmt = self._process_stmt(stmt, available, callback, pre_stmts, next_id, added)
            new_statements.extend(pre_stmts)
            new_statements.append(stmt)

        pre_stmts = []
        block.test = self._process_expr(block.test, available, callback, pre_stmts, next_id, added)
        new_statements.extend(pre_stmts)

        block.statements = new_statements

        for child in block.dom_children:
            self._process_block(child, available, callback, next_id)

        for expr in added:
            del available[expr]

    def _process_stmt(self, stmt, available, callback, pre_stmts, next_id, added):
        match stmt:
            case IRSet(place=SSAPlace() as place, value=IRPureInstr(op=op, args=args)) if self._is_cse_candidate(
                stmt.value, callback
            ):
                # Top-level CSE-candidate: process args but handle the outer expression
                # specially to avoid unnecessary extraction (recording is free here).
                new_args = [self._process_expr(arg, available, callback, pre_stmts, next_id, added) for arg in args]
                new_value = IRPureInstr(op=op, args=new_args)
                if new_value in available:
                    return IRSet(place, IRGet(available[new_value]))
                available[new_value] = place
                added.append(new_value)
                return IRSet(place, new_value)
            case IRSet(place=place, value=value):
                new_value = self._process_expr(value, available, callback, pre_stmts, next_id, added)
                return IRSet(place, new_value)
            case _:
                return self._process_expr(stmt, available, callback, pre_stmts, next_id, added)

    def _process_expr(self, expr, available, callback, pre_stmts, next_id, added):
        match expr:
            case IRPureInstr(op=op, args=args) if self._is_cse_candidate(expr, callback):
                new_args = [self._process_expr(arg, available, callback, pre_stmts, next_id, added) for arg in args]
                new_expr = IRPureInstr(op=op, args=new_args)
                if new_expr in available:
                    return IRGet(available[new_expr])
                if _cost(new_expr) > 4:
                    new_place = SSAPlace("_cse", next_id[0])
                    next_id[0] += 1
                    pre_stmts.append(IRSet(new_place, new_expr))
                    available[new_expr] = new_place
                    added.append(new_expr)
                    return IRGet(new_place)
                return new_expr
            case IRPureInstr(op=op, args=args) | IRInstr(op=op, args=args):
                new_args = [self._process_expr(arg, available, callback, pre_stmts, next_id, added) for arg in args]
                return type(expr)(op=op, args=new_args)
            case _:
                return expr



from pydori.lib.layer import LAYER_CONNECTOR, LAYER_SIM_LINE, get_z
from pydori.lib.options import Options
from pydori.lib.skin import Skin
from pydori.preview.layout import layout_preview_connector, layout_preview_sim_line, time_to_preview_col
from pydori.preview.note import PreviewNote
from sonolus.script.archetype import EntityRef, PreviewArchetype, imported


class PreviewHoldConnector(PreviewArchetype):
    """A connector for hold notes."""

    name = "HoldConnector"

    first_ref: EntityRef[PreviewNote] = imported()
    second_ref: EntityRef[PreviewNote] = imported()

    def render(self):
        first_col = time_to_preview_col(self.first.target_time)
        second_col = time_to_preview_col(self.second.target_time)
        for col in range(first_col, second_col + 1):
            Skin.hold_connector.draw(
                layout_preview_connector(
                    self.first.lane, self.second.lane, self.first.target_time, self.second.target_time, col
                ),
                z=get_z(
                    LAYER_CONNECTOR,
                    lane=min(self.first.lane, self.second.lane),
                    y=min(self.first.target_time, self.second.target_time),
                ),
                a=Options.connector_alpha,
            )

    @property
    def first(self):
        return self.first_ref.get()

    @property
    def second(self):
        return self.second_ref.get()


class PreviewSimLine(PreviewArchetype):
    """A line connecting two simultaneous notes."""

    name = "SimLine"

    first_ref: EntityRef[PreviewNote] = imported()
    second_ref: EntityRef[PreviewNote] = imported()

    def render(self):
        if Options.sim_lines_enabled:
            Skin.sim_line.draw(
                layout_preview_sim_line(self.first.lane, self.second.lane, self.first.target_time),
                z=get_z(
                    LAYER_SIM_LINE,
                    lane=min(self.first.lane, self.second.lane),
                    y=self.first.target_time,
                ),
                a=Options.sim_line_alpha,
            )

    @property
    def first(self):
        return self.first_ref.get()

    @property
    def second(self):
        return self.second_ref.get()



from hypothesis import assume, given
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.interval import Interval, interp, interp_clamped, lerp, remap, remap_clamped
from tests.script.conftest import implies, is_close, run_and_validate

ints = st.integers(min_value=-99999, max_value=99999)
floats = st.floats(min_value=-99999, max_value=99999, allow_infinity=False, allow_nan=False)
positive_deltas = st.floats(min_value=1e-4, max_value=999, allow_infinity=False, allow_nan=False)
floats_0_1 = st.floats(min_value=0, max_value=1, allow_infinity=False, allow_nan=False)
divisor_floats = floats.filter(lambda x: abs(x) > 1e-6)


@st.composite
def xp_fp_pairs(draw):
    size = draw(st.integers(min_value=2, max_value=10))
    xp_start = draw(floats)
    deltas = draw(st.lists(positive_deltas, min_size=size - 1, max_size=size - 1))
    fp_values = draw(st.lists(floats, min_size=size, max_size=size))

    xp_tuple = (xp_start, *tuple(xp_start + sum(deltas[: i + 1]) for i in range(len(deltas))))
    fp_tuple = tuple(fp_values)

    return xp_tuple, fp_tuple


@st.composite
def xp_fp_pairs_monotonic(draw):
    size = draw(st.integers(min_value=2, max_value=10))
    xp_start = draw(floats)
    fp_start = draw(floats)
    xp_deltas = draw(st.lists(positive_deltas, min_size=size - 1, max_size=size - 1))
    fp_deltas = draw(st.lists(positive_deltas, min_size=size - 1, max_size=size - 1))

    xp_tuple = (xp_start, *tuple(xp_start + sum(xp_deltas[: i + 1]) for i in range(len(xp_deltas))))
    fp_tuple = (fp_start, *tuple(fp_start + sum(fp_deltas[: i + 1]) for i in range(len(fp_deltas))))

    return xp_tuple, fp_tuple


@given(floats, floats)
def test_interval_length(start, end):
    def fn():
        interval = Interval(start, end)
        return interval.length

    assert run_and_validate(fn) == end - start


@given(floats, floats)
def test_interval_mid(start, end):
    def fn():
        interval = Interval(start, end)
        return interval.mid

    assert run_and_validate(fn) == (start + end) / 2


@given(floats, floats)
def test_interval_is_empty(start, end):
    def fn():
        interval = Interval(start, end)
        return interval.is_empty

    assert run_and_validate(fn) == (start > end)


@given(floats, floats, floats)
def test_interval_contains_float(start, end, value):
    def fn():
        interval = Interval(start, end)
        return value in interval

    assert run_and_validate(fn) == (start <= value <= end)


@given(floats, floats, floats, floats)
def test_interval_contains_interval(start, end, other_start, other_end):
    def fn():
        interval = Interval(start, end)
        other = Interval(other_start, other_end)
        return other in interval

    assert run_and_validate(fn) == (start <= other_start and other_end <= end)


@given(floats, floats, floats)
def test_interval_add(start, end, value):
    def fn():
        interval = Interval(start, end)
        return interval + value

    assert run_and_validate(fn) == Interval(start + value, end + value)


@given(floats, floats, floats)
def test_interval_sub(start, end, value):
    def fn():
        interval = Interval(start, end)
        return interval - value

    assert run_and_validate(fn) == Interval(start - value, end - value)


@given(floats, floats, floats)
def test_interval_mul(start, end, value):
    def fn():
        interval = Interval(start, end)
        return interval * value

    assert run_and_validate(fn) == Interval(start * value, end * value)


@given(floats, floats, divisor_floats)
def test_interval_truediv(start, end, value):
    def fn():
        interval = Interval(start, end)
        return interval / value

    assert run_and_validate(fn) == Interval(start / value, end / value)


@given(ints, ints, ints.filter(lambda x: x != 0))
def test_interval_floordiv(start, end, value):
    def fn():
        interval = Interval(start, end)
        return interval // value

    assert run_and_validate(fn) == Interval(start // value, end // value)


@given(floats, floats, floats, floats)
def test_interval_intersection(start, end, other_start, other_end):
    def fn():
        interval = Interval(start, end)
        other = Interval(other_start, other_end)
        return interval & other

    assert run_and_validate(fn) == Interval(max(start, other_start), min(end, other_end))


@given(floats, floats, floats, floats)
def test_interval_intersection_is_no_longer_than_original(start, end, other_start, other_end):
    def fn():
        interval = Interval(start, end)
        other = Interval(other_start, other_end)
        intersection = interval & other
        return intersection.length <= interval.length and intersection.length <= other.length

    assert run_and_validate(fn)


@given(floats, floats, floats, floats, floats)
def test_interval_intersection_consistent_with_contains(start, end, other_start, other_end, value):
    def fn():
        interval = Interval(start, end)
        other = Interval(other_start, other_end)
        return ((value in interval) and (value in other)) == (value in (interval & other))

    assert run_and_validate(fn)


@given(floats, floats, floats, floats, floats, floats)
def test_interval_transitive_contains(start1, end1, start2, end2, start3, end3):
    def fn():
        interval1 = Interval(start1, end1)
        interval2 = Interval(start2, end2)
        interval3 = Interval(start3, end3)
        return implies(interval1 in interval2 in interval3, interval1 in interval3)

    assert run_and_validate(fn)


@given(floats, floats, floats, floats)
def test_interval_intersection_is_in_original(start, end, other_start, other_end):
    def fn():
        interval = Interval(start, end)
        other = Interval(other_start, other_end)
        intersection = interval & other
        return intersection in interval and intersection in other

    assert run_and_validate(fn)


@given(floats, floats, floats, floats, floats)
def test_remap_inverse(start, end, other_start, other_end, value):
    assume(abs(end - start) > 1e-4 and abs(other_end - other_start) > 1e-4)

    def fn():
        remapped = remap(start, end, other_start, other_end, value)
        return remap(other_start, other_end, start, end, remapped)

    assert is_close(run_and_validate(fn), value, abs_tol=1e-3)


@given(floats, floats, floats, floats, floats)
def test_remap_clamped_inverse(start, end, other_start, other_end, value):
    assume(abs(end - start) > 1e-6 and abs(other_end - other_start) > 1e-6)

    def fn():
        remapped = remap_clamped(start, end, other_start, other_end, value)
        return remap_clamped(other_start, other_end, start, end, remapped)

    assert is_close(run_and_validate(fn), sorted([start, value, end])[1], abs_tol=1e-4)


@given(floats, floats, floats)
def test_lerp_unlerp_inverse(start, end, value):
    assume(abs(end - start) > 1e-6)

    def fn():
        interval = Interval(start, end)
        lerped = interval.lerp(value)
        return interval.unlerp(lerped)

    assert is_close(run_and_validate(fn), value, abs_tol=1e-4)


@given(floats, floats, floats)
def test_unlerp_lerp_inverse(start, end, value):
    assume(abs(end - start) > 1e-6)

    def fn():
        interval = Interval(start, end)
        unlerped = interval.unlerp(value)
        return interval.lerp(unlerped)

    assert is_close(run_and_validate(fn), value, abs_tol=1e-4)


@given(floats, floats, floats)
def test_lerp_clamped_unlerp_clamped_inverse(start, end, value):
    assume(abs(end - start) > 1e-6)

    def fn():
        interval = Interval(start, end)
        lerped_clamped = interval.lerp_clamped(value)
        return interval.unlerp_clamped(lerped_clamped)

    assert is_close(run_and_validate(fn), sorted([0, value, 1])[1], abs_tol=1e-4)


@given(floats, floats, floats)
def test_unlerp_clamped_lerp_clamped_inverse(start, end, value):
    assume(abs(end - start) > 1e-6)

    def fn():
        interval = Interval(start, end)
        unlerped_clamped = interval.unlerp_clamped(value)
        return interval.lerp_clamped(unlerped_clamped)

    assert is_close(run_and_validate(fn), sorted([start, value, end])[1], abs_tol=1e-4)


@given(xp_fp_pairs(), positive_deltas)
def test_interp_bounds_arrays(xp_fp_pair, offset):
    xp_tuple, fp_tuple = xp_fp_pair

    def fn():
        xp = Array(*xp_tuple)
        fp = Array(*fp_tuple)

        x_below = xp_tuple[0] - offset
        result_below = interp(xp, fp, x_below)
        expected_below = remap(xp_tuple[0], xp_tuple[1], fp_tuple[0], fp_tuple[1], x_below)

        x_above = xp_tuple[-1] + offset
        result_above = interp(xp, fp, x_above)
        expected_above = remap(xp_tuple[-2], xp_tuple[-1], fp_tuple[-2], fp_tuple[-1], x_above)

        return Array(result_below, expected_below, result_above, expected_above)

    result_below, expected_below, result_above, expected_above = run_and_validate(fn)
    assert is_close(result_below, expected_below, abs_tol=1e-4)
    assert is_close(result_above, expected_above, abs_tol=1e-4)


@given(xp_fp_pairs(), positive_deltas)
def test_interp_bounds_tuples(xp_fp_pair, offset):
    xp_tuple, fp_tuple = xp_fp_pair

    def fn():
        xp = xp_tuple
        fp = fp_tuple

        x_below = xp_tuple[0] - offset
        result_below = interp(xp, fp, x_below)
        expected_below = remap(xp_tuple[0], xp_tuple[1], fp_tuple[0], fp_tuple[1], x_below)

        x_above = xp_tuple[-1] + offset
        result_above = interp(xp, fp, x_above)
        expected_above = remap(xp_tuple[-2], xp_tuple[-1], fp_tuple[-2], fp_tuple[-1], x_above)

        return Array(result_below, expected_below, result_above, expected_above)

    result_below, expected_below, result_above, expected_above = run_and_validate(fn)
    assert is_close(result_below, expected_below, abs_tol=1e-4)
    assert is_close(result_above, expected_above, abs_tol=1e-4)


@given(xp_fp_pairs(), positive_deltas.filter(lambda x: abs(x) > 1e-6))
def test_interp_clamped_bounds_arrays(xp_fp_pair, offset):
    xp_tuple, fp_tuple = xp_fp_pair

    def fn():
        xp = Array(*xp_tuple)
        fp = Array(*fp_tuple)

        x_below = xp_tuple[0] - offset
        result_below = interp_clamped(xp, fp, x_below)

        x_above = xp_tuple[-1] + offset
        result_above = interp_clamped(xp, fp, x_above)

        return Array(result_below, fp_tuple[0], result_above, fp_tuple[-1])

    result_below, expected_below, result_above, expected_above = run_and_validate(fn)
    assert is_close(result_below, expected_below, abs_tol=1e-4)
    assert is_close(result_above, expected_above, abs_tol=1e-4)


@given(xp_fp_pairs(), positive_deltas)
def test_interp_clamped_bounds_tuples(xp_fp_pair, offset):
    xp_tuple, fp_tuple = xp_fp_pair

    def fn():
        xp = xp_tuple
        fp = fp_tuple

        x_below = xp_tuple[0] - offset
        result_below = interp_clamped(xp, fp, x_below)

        x_above = xp_tuple[-1] + offset
        result_above = interp_clamped(xp, fp, x_above)

        return Array(result_below, fp_tuple[0], result_above, fp_tuple[-1])

    result_below, expected_below, result_above, expected_above = run_and_validate(fn)
    assert is_close(result_below, expected_below, abs_tol=1e-4)
    assert is_close(result_above, expected_above, abs_tol=1e-4)


@given(xp_fp_pairs(), floats_0_1)
def test_interp_within_bounds_arrays(xp_fp_pair, rel_x):
    xp_tuple, fp_tuple = xp_fp_pair

    x = lerp(xp_tuple[0], xp_tuple[-1], rel_x)

    def fn():
        xp = Array(*xp_tuple)
        fp = Array(*fp_tuple)
        result = interp(xp, fp, x)
        return Array(result, min(fp_tuple), max(fp_tuple))

    result, min_fp, max_fp = run_and_validate(fn)
    assert (
        min_fp <= result <= max_fp or is_close(result, min_fp, abs_tol=1e-4) or is_close(result, max_fp, abs_tol=1e-4)
    )


@given(xp_fp_pairs(), floats_0_1)
def test_interp_within_bounds_tuples(xp_fp_pair, rel_x):
    xp_tuple, fp_tuple = xp_fp_pair

    x = lerp(xp_tuple[0], xp_tuple[-1], rel_x)

    def fn():
        xp = xp_tuple
        fp = fp_tuple
        result = interp(xp, fp, x)
        return Array(result, min(fp_tuple), max(fp_tuple))

    result, min_fp, max_fp = run_and_validate(fn)
    assert (
        min_fp <= result <= max_fp or is_close(result, min_fp, abs_tol=1e-4) or is_close(result, max_fp, abs_tol=1e-4)
    )


@given(xp_fp_pairs(), floats_0_1)
def test_interp_clamped_within_bounds_arrays(xp_fp_pair, rel_x):
    xp_tuple, fp_tuple = xp_fp_pair

    x = lerp(xp_tuple[0], xp_tuple[-1], rel_x)

    def fn():
        xp = Array(*xp_tuple)
        fp = Array(*fp_tuple)
        result = interp_clamped(xp, fp, x)
        return Array(result, min(fp_tuple), max(fp_tuple))

    result, min_fp, max_fp = run_and_validate(fn)
    assert (
        min_fp <= result <= max_fp or is_close(result, min_fp, abs_tol=1e-4) or is_close(result, max_fp, abs_tol=1e-4)
    )


@given(xp_fp_pairs(), floats)
def test_interp_clamped_within_bounds_tuples(xp_fp_pair, rel_x):
    xp_tuple, fp_tuple = xp_fp_pair

    # Ensure x is within bounds
    x = xp_tuple[0] + (xp_tuple[-1] - xp_tuple[0]) * max(0, min(1, (rel_x + 1) / 2))

    def fn():
        xp = xp_tuple
        fp = fp_tuple
        result = interp_clamped(xp, fp, x)
        return Array(result, min(fp_tuple), max(fp_tuple))

    result, min_fp, max_fp = run_and_validate(fn)
    assert (
        min_fp <= result <= max_fp or is_close(result, min_fp, abs_tol=1e-4) or is_close(result, max_fp, abs_tol=1e-4)
    )


def test_interp_simple_cases():
    def fn():
        # Test middle point
        xp = Array(0.0, 10.0)
        fp = Array(5.0, 15.0)
        middle = interp(xp, fp, 5.0)

        # Test exact point match
        exact = interp(xp, fp, 0.0)

        # Test outside bounds
        outside = interp(xp, fp, 20.0)

        return Array(middle, exact, outside)

    middle, exact, outside = run_and_validate(fn)
    assert is_close(middle, 10.0, abs_tol=1e-4)  # Midpoint between 5 and 15
    assert is_close(exact, 5.0, abs_tol=1e-4)  # Exact match at x=0
    assert is_close(outside, 25.0, abs_tol=1e-4)  # Extrapolation: 15 + (15-5) * (20-10)/(10-0)


def test_interp_clamped_simple_cases():
    def fn():
        # Test middle point
        xp = Array(0.0, 10.0)
        fp = Array(5.0, 15.0)
        middle = interp_clamped(xp, fp, 5.0)

        # Test exact point match
        exact = interp_clamped(xp, fp, 0.0)

        # Test outside bounds (should clamp)
        outside_low = interp_clamped(xp, fp, -5.0)
        outside_high = interp_clamped(xp, fp, 20.0)

        return Array(middle, exact, outside_low, outside_high)

    middle, exact, outside_low, outside_high = run_and_validate(fn)
    assert is_close(middle, 10.0, abs_tol=1e-4)  # Midpoint between 5 and 15
    assert is_close(exact, 5.0, abs_tol=1e-4)  # Exact match at x=0
    assert is_close(outside_low, 5.0, abs_tol=1e-4)  # Clamped to first value
    assert is_close(outside_high, 15.0, abs_tol=1e-4)  # Clamped to last value


def test_interp_tuple_simple_cases():
    def fn():
        # Test middle point
        xp = (0.0, 10.0)
        fp = (5.0, 15.0)
        middle = interp(xp, fp, 5.0)

        # Test exact point match
        exact = interp(xp, fp, 0.0)

        # Test outside bounds
        outside = interp(xp, fp, 20.0)

        return Array(middle, exact, outside)

    middle, exact, outside = run_and_validate(fn)
    assert is_close(middle, 10.0, abs_tol=1e-4)  # Midpoint between 5 and 15
    assert is_close(exact, 5.0, abs_tol=1e-4)  # Exact match at x=0
    assert is_close(outside, 25.0, abs_tol=1e-4)  # Extrapolation: 15 + (15-5) * (20-10)/(10-0)


def test_interp_clamped_tuple_simple_cases():
    def fn():
        # Test middle point
        xp = (0.0, 10.0)
        fp = (5.0, 15.0)
        middle = interp_clamped(xp, fp, 5.0)

        # Test exact point match
        exact = interp_clamped(xp, fp, 0.0)

        # Test outside bounds (should clamp)
        outside_low = interp_clamped(xp, fp, -5.0)
        outside_high = interp_clamped(xp, fp, 20.0)

        return Array(middle, exact, outside_low, outside_high)

    middle, exact, outside_low, outside_high = run_and_validate(fn)
    assert is_close(middle, 10.0, abs_tol=1e-4)  # Midpoint between 5 and 15
    assert is_close(exact, 5.0, abs_tol=1e-4)  # Exact match at x=0
    assert is_close(outside_low, 5.0, abs_tol=1e-4)  # Clamped to first value
    assert is_close(outside_high, 15.0, abs_tol=1e-4)  # Clamped to last value


@given(xp_fp_pairs_monotonic(), floats)
def test_interp_inverse_arrays(xp_fp_pair, x):
    xp_tuple, fp_tuple = xp_fp_pair

    def fn():
        xp = Array(*xp_tuple)
        fp = Array(*fp_tuple)

        y = interp(xp, fp, x)
        x_recovered = interp(fp, xp, y)

        return Array(x, x_recovered)

    original_x, recovered_x = run_and_validate(fn)
    assert is_close(original_x, recovered_x, abs_tol=1e-3)


@given(xp_fp_pairs_monotonic(), floats)
def test_interp_inverse_tuples(xp_fp_pair, x):
    xp_tuple, fp_tuple = xp_fp_pair

    def fn():
        xp = xp_tuple
        fp = fp_tuple

        y = interp(xp, fp, x)
        x_recovered = interp(fp, xp, y)

        return Array(x, x_recovered)

    original_x, recovered_x = run_and_validate(fn)
    assert is_close(original_x, recovered_x, abs_tol=1e-3)


@given(xp_fp_pairs_monotonic(), floats_0_1)
def test_interp_clamped_inverse_arrays(xp_fp_pair, rel_x):
    xp_tuple, fp_tuple = xp_fp_pair

    x = lerp(xp_tuple[0], xp_tuple[-1], rel_x)

    def fn():
        xp = Array(*xp_tuple)
        fp = Array(*fp_tuple)

        y = interp_clamped(xp, fp, x)
        x_recovered = interp_clamped(fp, xp, y)

        return Array(x, x_recovered)

    original_x, recovered_x = run_and_validate(fn)
    assert is_close(original_x, recovered_x, abs_tol=1e-3)


@given(xp_fp_pairs_monotonic(), floats_0_1)
def test_interp_clamped_inverse_tuples(xp_fp_pair, rel_x):
    xp_tuple, fp_tuple = xp_fp_pair

    x = lerp(xp_tuple[0], xp_tuple[-1], rel_x)

    def fn():
        xp = xp_tuple
        fp = fp_tuple

        y = interp_clamped(xp, fp, x)
        x_recovered = interp_clamped(fp, xp, y)

        return Array(x, x_recovered)

    original_x, recovered_x = run_and_validate(fn)
    assert is_close(original_x, recovered_x, abs_tol=1e-3)



from pydori.lib.layer import LAYER_BPM_CHANGE_LINE, LAYER_MEASURE_LINE, LAYER_TIMESCALE_CHANGE_LINE, get_z
from pydori.lib.skin import Skin
from pydori.preview.layout import PREVIEW_BAR_LINE_ALPHA, PreviewData, layout_preview_bar_line, print_at_time
from sonolus.script.archetype import PreviewArchetype, StandardArchetypeName, StandardImport, entity_data, imported
from sonolus.script.printing import PrintColor, PrintFormat
from sonolus.script.timing import beat_to_starting_beat, beat_to_time

# Number of beats between measure lines drawn in the preview.
METER = 4


class PreviewBpmChange(PreviewArchetype):
    name = StandardArchetypeName.BPM_CHANGE

    beat: StandardImport.BEAT = imported()
    bpm: StandardImport.BPM = imported()

    time: float = entity_data()

    def preprocess(self):
        self.time = beat_to_time(self.beat)

    def render(self):
        self.render_bpm_line()
        self.render_succeeding_measure_lines()

    def render_bpm_line(self):
        Skin.bpm_change_line.draw(
            layout_preview_bar_line(self.time, "right"),
            z=get_z(LAYER_BPM_CHANGE_LINE),
            a=PREVIEW_BAR_LINE_ALPHA,
        )
        print_at_time(
            self.bpm,
            self.time,
            fmt=PrintFormat.BPM,
            color=PrintColor.PURPLE,
            side="right",
        )

    def render_succeeding_measure_lines(self):
        beat = self.beat + METER
        while beat_to_starting_beat(beat) == self.beat and beat <= PreviewData.last_beat:
            Skin.measure_line.draw(
                layout_preview_bar_line(beat_to_time(beat), "none"),
                z=LAYER_MEASURE_LINE - beat_to_time(beat) / 100,
                a=PREVIEW_BAR_LINE_ALPHA,
            )
            beat += METER


class PreviewTimescaleChange(PreviewArchetype):
    name = StandardArchetypeName.TIMESCALE_CHANGE

    beat: StandardImport.BEAT = imported()
    timescale: StandardImport.TIMESCALE = imported()

    time: float = entity_data()

    def preprocess(self):
        self.time = beat_to_time(self.beat)

    def render(self):
        Skin.timescale_change_line.draw(
            layout_preview_bar_line(self.time, "left"),
            z=get_z(LAYER_TIMESCALE_CHANGE_LINE),
            a=PREVIEW_BAR_LINE_ALPHA,
        )
        print_at_time(
            self.timescale,
            self.time,
            fmt=PrintFormat.TIMESCALE,
            color=PrintColor.YELLOW,
            side="left",
        )



# ruff: noqa: B905
from enum import Enum
from typing import Any, Self

from sonolus.script.internal.impl import validate_value
from sonolus.script.internal.simple_meta_fn import simple_meta_fn
from sonolus.script.internal.transient import TransientValue


class TupleImpl(TransientValue):
    value: tuple

    def __init__(self, value: tuple):
        self.value = value

    @simple_meta_fn
    def __getitem__(self, item):
        item = validate_value(item)
        if not item._is_py_():
            raise TypeError(f"Cannot index tuple with non compile-time constant {item}")
        item = item._as_py_()
        if not isinstance(item, int | float):
            raise TypeError(f"Cannot index tuple with {item}")
        if int(item) != item:
            raise TypeError(f"Cannot index tuple with non-integer {item}")
        if not (-len(self.value) <= item < len(self.value)):
            raise IndexError(f"Tuple index out of range: {item}")
        if item < 0:
            item += len(self.value)
        return self.value[int(item)]

    @simple_meta_fn
    def __len__(self):
        return len(self.value)

    def __eq__(self, other):
        if not self._is_tuple_impl(other):
            return False
        if len(self) != len(other):
            return False
        for a, b in zip(self, other):  # noqa: SIM110
            if a != b:
                return False
        return True

    def __ne__(self, other):
        if not self._is_tuple_impl(other):
            return True
        if len(self) != len(other):
            return True
        for a, b in zip(self, other):  # noqa: SIM110
            if a != b:
                return True
        return False

    def __lt__(self, other):
        if not self._is_tuple_impl(other):
            return NotImplemented
        for a, b in zip(self, other):
            if a != b:
                return a < b
        return len(self.value) < len(other.value)

    def __le__(self, other):
        if not self._is_tuple_impl(other):
            return NotImplemented
        for a, b in zip(self, other):
            if a != b:
                return a < b
        return len(self.value) <= len(other.value)

    def __gt__(self, other):
        if not self._is_tuple_impl(other):
            return NotImplemented
        for a, b in zip(self, other):
            if a != b:
                return a > b
        return len(self.value) > len(other.value)

    def __ge__(self, other):
        if not self._is_tuple_impl(other):
            return NotImplemented
        for a, b in zip(self, other):
            if a != b:
                return a > b
        return len(self.value) >= len(other.value)

    def __hash__(self):
        return hash(self.value)

    @simple_meta_fn
    def __add__(self, other) -> Self:
        other = TupleImpl._accept_(other)
        return TupleImpl._accept_(self.value + other.value)

    def __contains__(self, item):
        for element in self.value:  # noqa: SIM110
            if element == item:
                return True
        return False

    @staticmethod
    @simple_meta_fn
    def _is_tuple_impl(value: Any) -> bool:
        return isinstance(value, TupleImpl)

    @classmethod
    def _accepts_(cls, value: Any) -> bool:
        return isinstance(value, cls | tuple)

    @classmethod
    def _accept_(cls, value: Any) -> Self:
        if not cls._accepts_(value):
            raise TypeError(f"Cannot accept {value} as {cls.__name__}")
        if isinstance(value, cls):
            return value
        else:
            return cls(tuple(validate_value(item) for item in value))

    def _is_py_(self) -> bool:
        return all(item._is_py_() for item in self.value)

    def _as_py_(self) -> tuple:
        return tuple(item._as_py_() for item in self.value)

    def _tuple_iter_(self) -> tuple:
        return self.value


TupleImpl.__name__ = "tuple"
TupleImpl.__qualname__ = "tuple"


def has_tuple_iter(v) -> bool:
    v = validate_value(v)
    if hasattr(v, "_tuple_iter_"):
        return True
    if v._is_py_():
        v_py = v._as_py_()
        if isinstance(v_py, type):
            return issubclass(v_py, Enum)
    return False


def tuple_iter(v) -> tuple:
    v = validate_value(v)
    if hasattr(v, "_tuple_iter_"):
        return v._tuple_iter_()
    if v._is_py_():
        v_py = v._as_py_()
        if isinstance(v_py, type):
            if issubclass(v_py, Enum):
                return tuple(v_py)
            raise TypeError(f"Cannot iterate over non-Enum class {v_py}")
    raise TypeError(f"Cannot iterate over {v}")



from pydori.tutorial.update import dec_phase, inc_phase
from sonolus.script.runtime import navigation_direction


def navigate():
    if navigation_direction() > 0:
        inc_phase()
    else:
        dec_phase()



import pytest

from sonolus.script.array import Array
from sonolus.script.debug import debug_log
from sonolus.script.internal.error import CompilationError
from sonolus.script.record import Record
from tests.script.conftest import run_and_validate, run_compiled


class UnsupportedDescriptor:
    def __get__(self, instance, owner):
        return -1

    def __set__(self, instance, value):
        pass


class MyBox[T](Record):
    value: T

    @classmethod
    def my_classmethod(cls):
        return 123

    def my_method(self):
        return 456

    @property
    def my_property(self):
        return 789

    @my_property.setter
    def my_property(self, value):
        debug_log(value)


MyBox.unsupported_descriptor = UnsupportedDescriptor()


def test_hasattr_record_field():
    def fn():
        return hasattr(MyBox(1), "value")

    assert run_and_validate(fn) == 1


def test_hasattr_record_method():
    def fn():
        return hasattr(MyBox(1), "my_method")

    assert run_and_validate(fn) == 1


def test_hasattr_record_classmethod():
    def fn():
        return hasattr(MyBox(1), "my_classmethod")

    assert run_and_validate(fn) == 1


def test_hasattr_record_property():
    def fn():
        return hasattr(MyBox(1), "my_property")

    assert run_and_validate(fn) == 1


def test_hasattr_record_not_present():
    def fn():
        return hasattr(MyBox(1), "does_not_exist")

    assert run_and_validate(fn) == 0


def test_hasattr_record_unsupported():
    def fn():
        return hasattr(MyBox(1), "unsupported_descriptor")

    with pytest.raises(CompilationError, match="Unsupported field"):
        run_compiled(fn)


def test_hasattr_array_method():
    def fn():
        return hasattr(Array(1), "__getitem__")

    assert run_and_validate(fn) == 1


def test_hasattr_array_not_present():
    def fn():
        return hasattr(Array(1), "does_not_exist")

    assert run_and_validate(fn) == 0


def test_hasattr_type_classmethod():
    def fn():
        return hasattr(MyBox, "my_classmethod")

    assert run_and_validate(fn) == 1


def test_hasattr_type_method():
    def fn():
        return hasattr(MyBox, "my_method")

    assert run_and_validate(fn) == 1


def test_hasattr_type_not_present():
    def fn():
        return hasattr(MyBox, "does_not_exist")

    assert run_and_validate(fn) == 0


def test_getattr_record_field():
    def fn():
        box = MyBox(100)
        return box.value

    assert run_and_validate(fn) == 100


def test_getattr_record_method():
    def fn():
        box = MyBox(100)
        method = box.my_method
        return method()

    assert run_and_validate(fn) == 456


def test_getattr_record_classmethod():
    def fn():
        method = MyBox.my_classmethod
        return method()

    assert run_and_validate(fn) == 123


def test_getattr_record_property():
    def fn():
        box = MyBox(100)
        return box.my_property

    assert run_and_validate(fn) == 789


def test_getattr_record_unsupported():
    def fn():
        box = MyBox(100)
        return box.unsupported_descriptor

    with pytest.raises(CompilationError, match="Unsupported field"):
        run_compiled(fn)


def test_getattr_array_method():
    def fn():
        arr = Array(10, 20, 30)
        method = arr.__getitem__
        return method(1)

    assert run_and_validate(fn) == 20


def test_getattr_type_classmethod():
    def fn():
        method = MyBox.my_classmethod
        return method()

    assert run_and_validate(fn) == 123


def test_getattr_type_method():
    def fn():
        method = MyBox.my_method
        return method(MyBox(0))

    assert run_and_validate(fn) == 456


def test_getattr_type_not_present():
    def fn():
        return MyBox.does_not_exist

    with pytest.raises(CompilationError, match="has no attribute 'does_not_exist'"):
        run_compiled(fn)


def test_setattr_record_field():
    def fn():
        box = MyBox(0)
        box.value = 100
        return box.value

    assert run_and_validate(fn) == 100


def test_setattr_record_property():
    def fn():
        box = MyBox(0)
        box.my_property = 100
        return 1

    assert run_and_validate(fn) == 1


def test_setattr_record_unsupported():
    def fn():
        box = MyBox(0)
        box.unsupported_descriptor = 100
        return 1

    with pytest.raises(CompilationError, match="Unsupported field"):
        run_compiled(fn)


def test_setattr_type_not_supported():
    def fn():
        MyBox.my_classmethod = 100
        return 1

    with pytest.raises(CompilationError, match="Unsupported field"):
        run_compiled(fn)



from typing import Any

from sonolus.script.array import Array
from sonolus.script.debug import error
from sonolus.script.internal.context import Context, ctx, set_ctx
from sonolus.script.internal.impl import validate_value
from sonolus.script.internal.meta_fn import meta_fn
from sonolus.script.maybe import Nothing, Some
from sonolus.script.num import Num, _is_num
from sonolus.script.record import Record


class DictImpl[Keys, OrderedKeys, Values](Record):
    _keys: Keys  # tuple[K, ...]
    _ordered_keys: OrderedKeys  # tuple[tuple[K, int], ...] | None
    _values: Values  # tuple[V, ...] | Array[V, ...]

    @property
    @meta_fn
    def _has_ordered_keys(self) -> bool:
        keys = validate_value(self._ordered_keys)
        return not (keys._is_py_() and keys._as_py_() is None)

    @property
    @meta_fn
    def _has_array_values(self) -> bool:
        values = validate_value(self._values)
        return isinstance(values, Array)

    @property
    @meta_fn
    def _size(self) -> int:
        return validate_value(len(self._keys))._as_py_()

    def __len__(self) -> int:
        return len(self._values)

    def __contains__(self, item):
        constsearch_res = self._try_constsearch(item)
        if constsearch_res is not None:
            return constsearch_res >= 0
        numsearch_res = self._try_numsearch(item)
        if numsearch_res is not None:
            return numsearch_res >= 0
        if self._has_ordered_keys:
            return self._binsearch(item) >= 0
        return self._linsearch(item) >= 0

    def __getitem__(self, item):
        result = self._maybe_getitem(item)
        return result.get(error_message="KeyError")

    def get(self, item, default, /):
        return self._maybe_getitem(item).or_default(default)

    def _maybe_getitem(self, item):
        index = self._try_constsearch(item)
        if index is not None:
            if index < 0:
                return Nothing
            elif isinstance(self._values, Array):
                return Some(self._values.get_unchecked(index))
            else:
                return Some(self._values[index])
        index = self._try_numsearch(item)
        if index is not None:
            if index < 0:
                return Nothing
            elif isinstance(self._values, Array):
                return Some(self._values.get_unchecked(index))
            else:
                return Some(self._values[index])
        if not self._has_array_values:
            error(
                "Dict must be accessed via a compile time constant unless "
                "all value are compile time constants of a uniform type."
            )
        if self._has_ordered_keys:
            index = self._binsearch(item)
        else:
            index = self._linsearch(item)
        if index < 0:
            return Nothing
        elif isinstance(self._values, Array):
            return Some(self._values.get_unchecked(index))
        else:
            return Some(self._values[index])

    def __eq__(self, other: Any):
        raise TypeError("Dict equality comparison is not supported")

    __hash__ = None

    @meta_fn
    def __or__(self, other):
        if not isinstance(other, DictImpl):
            raise TypeError("Unsupported type for '|' operator")
        return self.from_dict({**self._as_dict_with_py_keys(), **other._as_dict_with_py_keys()})

    @staticmethod
    def from_dict(d):
        keys = tuple(validate_value(k) for k in d)
        values = tuple(validate_value(v) for v in d.values())
        if not all(k._is_py_() for k in keys):
            raise TypeError("Dict keys must be a compile-time constant")
        if len(keys) >= 2:
            py_keys = [k._as_py_() for k in keys]
            py_key_types = {type(k) for k in py_keys}
            is_comparable = (
                len(py_key_types) == 1
                and py_keys[0].__lt__(py_keys[1]) is not NotImplemented  # noqa: PLC2801
                and type(keys[0]).__lt__ is not object.__lt__
            )
            if is_comparable:
                ordered_keys = tuple((k, i) for i, k in sorted(enumerate(py_keys), key=lambda x: x[1]))
            else:
                ordered_keys = None
        else:
            ordered_keys = None
        if all(v._is_py_() for v in values) and len({type(v) for v in values}) == 1:
            values = Array[type(values[0]), len(values)]._with_value([*values])
        return DictImpl(keys, ordered_keys, values)

    def _as_dict_with_py_keys(self):
        return {
            self._keys[i]._as_py_(): self._values._value[i]
            if isinstance(self._values, Array) and isinstance(self._values._value, list)
            else self._values[i]
            for i in range(self._size)
        }

    @meta_fn
    def _try_constsearch(self, item):
        from sonolus.script.internal.visitor import compile_and_call

        orig_ctx = ctx()
        begin_ctx = orig_ctx.branch(None)
        set_ctx(begin_ctx)

        for i, k in enumerate(self._keys):
            eq = validate_value(compile_and_call(k.__eq__, item))
            if not eq._is_py_():
                # A bit of a hack to allow resetting to the original state
                del orig_ctx.outgoing[None]
                set_ctx(orig_ctx)
                return None
            if eq._as_py_():
                return i
        return -1

    @meta_fn
    def _try_numsearch(self, item):
        item = validate_value(item)
        if not _is_num(item):
            return None
        res = Num._alloc_()
        orig_ctx = ctx()
        begin_ctx = orig_ctx.branch(None)
        set_ctx(begin_ctx)
        begin_ctx.test = item.ir()
        end_ctxs = []
        for i, k in enumerate(self._keys):
            if not _is_num(k):
                # A bit of a hack to allow resetting to the original state
                del orig_ctx.outgoing[None]
                set_ctx(orig_ctx)
                return None
            set_ctx(begin_ctx.branch(k._as_py_()))
            res._set_(i)
            end_ctxs.append(ctx())
        set_ctx(begin_ctx.branch(None))
        res._set_(-1)
        end_ctxs.append(ctx())
        set_ctx(Context.meet(end_ctxs))
        return res

    @meta_fn
    def _binsearch(self, item):
        return self._binsearch_internal(item, 0, self._size, Num._alloc_())

    @meta_fn
    def _binsearch_internal(self, item, lo, hi, res):
        from sonolus.script.internal.visitor import compile_and_call

        if lo >= hi:
            res._set_(-1)
            return res

        if hi - lo <= 3:
            # Linear search
            lo_value, orig_index = self._ordered_keys[lo]
            eq_test = compile_and_call(item.__eq__, lo_value).ir()
            ctx_init = ctx()
            ctx_init.test = eq_test
            eq_ctx = ctx_init.branch(None)
            neq_ctx = ctx_init.branch(0)

            set_ctx(eq_ctx)
            res._set_(orig_index)
            after_eq_ctx = ctx()

            set_ctx(neq_ctx)
            self._binsearch_internal(item, lo + 1, hi, res)

            set_ctx(Context.meet([after_eq_ctx, ctx()]))
            return res

        mid = (lo + hi) // 2
        mid_value, orig_index = self._ordered_keys[mid]
        eq_test = compile_and_call(item.__eq__, mid_value).ir()
        ctx_init = ctx()
        ctx_init.test = eq_test
        eq_ctx = ctx_init.branch(None)
        neq_ctx = ctx_init.branch(0)

        set_ctx(eq_ctx)
        res._set_(orig_index)
        after_eq_ctx = ctx()

        set_ctx(neq_ctx)

        neq_test = compile_and_call(item.__lt__, mid_value).ir()
        neq_ctx = ctx()
        neq_ctx.test = neq_test
        lt_ctx = neq_ctx.branch(None)
        gt_ctx = neq_ctx.branch(0)

        set_ctx(lt_ctx)
        self._binsearch_internal(item, lo, mid, res)
        after_lt_ctx = ctx()
        set_ctx(gt_ctx)
        self._binsearch_internal(item, mid + 1, hi, res)
        after_gt_ctx = ctx()

        set_ctx(Context.meet([after_eq_ctx, after_gt_ctx, after_lt_ctx]))
        return res

    def _linsearch(self, item):
        for i, k in enumerate(self._keys):
            if k == item:
                return i
        return -1

    def _tuple_iter_(self) -> tuple:
        from sonolus.script.internal.tuple_impl import tuple_iter

        return tuple_iter(self._keys)

    @meta_fn
    def keys(self):
        return validate_value(tuple(self._keys[i] for i in range(self._size)))

    @meta_fn
    def values(self):
        return validate_value(
            tuple(
                self._values._value[i]
                if isinstance(self._values, Array) and isinstance(self._values._value, list)
                else self._values[i]
                for i in range(self._size)
            )
        )

    @meta_fn
    def items(self):
        return validate_value(
            tuple(
                (
                    self._keys[i],
                    self._values._value[i]
                    if isinstance(self._values, Array) and isinstance(self._values._value, list)
                    else self._values[i],
                )
                for i in range(self._size)
            )
        )



import inspect
from abc import ABC
from collections.abc import Sequence
from typing import Annotated

_missing = object()


def get_field_specifiers(
    cls,
    *,
    skip: frozenset[str] | set[str] = frozenset(),
    globals=None,  # noqa: A002
    locals=None,  # noqa: A002
    eval_str=True,
    included_classes: Sequence[type] | None = None,
):
    """Like inspect.get_annotations, but also turns class attributes into Annotated."""
    if included_classes is not None:
        results = {}
        for entry in reversed(included_classes):
            results.update(inspect.get_annotations(entry, eval_str=eval_str))
    else:
        results = inspect.get_annotations(cls, globals=globals, locals=locals, eval_str=eval_str)
    for key, value in results.items():
        if key in skip:
            continue
        class_value = getattr(cls, key, _missing)
        if class_value is not _missing and key not in skip:
            results[key] = Annotated[value, class_value]
    for key, value in cls.__dict__.items():
        if (
            key not in results
            and key not in skip
            and not key.startswith("__")
            and not callable(value)
            and not hasattr(value, "__func__")
            and not isinstance(value, property)
            and not (issubclass(cls, ABC) and (hasattr(ABC, key)))
        ):
            raise ValueError(f"Missing annotation for {cls.__name__}.{key}")
    for skipped_key in skip:
        if skipped_key in results:
            del results[skipped_key]
    return results



from sonolus.script.sprite import RenderMode, StandardSprite, skin


@skin
class Skin:
    # Since this engine uses accurate 3D perspective rendering, it should always use lightweight rendering.
    render_mode = RenderMode.LIGHTWEIGHT

    cover: StandardSprite.STAGE_COVER

    lane: StandardSprite.LANE
    stage_left_border: StandardSprite.STAGE_LEFT_BORDER
    stage_right_border: StandardSprite.STAGE_RIGHT_BORDER
    stage_middle: StandardSprite.STAGE_MIDDLE
    judgment_line: StandardSprite.JUDGMENT_LINE
    slot: StandardSprite.NOTE_SLOT

    tap_note: StandardSprite.NOTE_HEAD_CYAN

    flick_note: StandardSprite.NOTE_HEAD_RED
    flick_arrow: StandardSprite.DIRECTIONAL_MARKER_RED

    right_flick_note: StandardSprite.NOTE_HEAD_YELLOW
    right_flick_arrow: StandardSprite.DIRECTIONAL_MARKER_YELLOW

    left_flick_note: StandardSprite.NOTE_HEAD_PURPLE
    left_flick_arrow: StandardSprite.DIRECTIONAL_MARKER_PURPLE

    hold_head_note: StandardSprite.NOTE_HEAD_GREEN
    hold_end_note: StandardSprite.NOTE_TAIL_GREEN
    hold_tick_note: StandardSprite.NOTE_TICK_GREEN
    hold_connector: StandardSprite.NOTE_CONNECTION_GREEN_SEAMLESS

    sim_line: StandardSprite.SIMULTANEOUS_CONNECTION_NEUTRAL_SEAMLESS

    # Preview mode bar lines
    bpm_change_line: StandardSprite.GRID_PURPLE
    timescale_change_line: StandardSprite.GRID_YELLOW
    measure_line: StandardSprite.GRID_NEUTRAL
    time_line: StandardSprite.GRID_CYAN



from __future__ import annotations

from typing import cast

from pydori.lib.buckets import note_judgment_window
from pydori.lib.layout import Hitbox, get_note_y, preempt_time
from pydori.lib.note import (
    NoteKind,
    destroy_particle,
    draw_note,
    draw_note_head,
    get_flick_speed_threshold,
    get_note_bucket,
    init_note_life,
    play_note_particle,
    play_note_sfx,
    schedule_note_sfx,
    stop_looped_sfx,
    update_hold_particle,
    update_hold_sfx,
)
from pydori.lib.options import Options
from pydori.lib.streams import Streams
from pydori.play.input import claim_touch, unclaimed_taps, unclaimed_touches
from sonolus.script.archetype import (
    EntityRef,
    PlayArchetype,
    StandardImport,
    callback,
    entity_data,
    entity_memory,
    exported,
    imported,
    shared_memory,
)
from sonolus.script.array import Dim
from sonolus.script.bucket import Judgment, JudgmentWindow
from sonolus.script.containers import VarArray
from sonolus.script.effect import LoopedEffectHandle
from sonolus.script.globals import level_memory
from sonolus.script.interval import Interval, clamp
from sonolus.script.particle import ParticleHandle
from sonolus.script.quad import Quad
from sonolus.script.runtime import input_offset, offset_adjusted_time, scaled_time, time, touches
from sonolus.script.timing import beat_to_time, time_to_scaled_time
from sonolus.script.vec import Vec2

DEFAULT_BEST_JUDGMENT_TIME = -1e8


class Note(PlayArchetype):
    """Common archetype for notes."""

    lane: float = imported()
    beat: StandardImport.BEAT = imported()
    direction: int = imported()
    prev_ref: EntityRef[Note] = imported()
    next_ref: EntityRef[Note] = imported()

    judgment_window: JudgmentWindow = entity_data()
    target_time: float = entity_data()
    target_scaled_time: float = entity_data()
    start_scaled_time: float = entity_data()
    input_interval: Interval = entity_data()
    head_ref: EntityRef[Note] = entity_data()
    end_ref: EntityRef[Note] = entity_data()

    best_judgment_time: float = entity_memory()

    _active_touch_id: int = shared_memory()
    _hold_lane: float = shared_memory()
    is_judged: bool = shared_memory()

    end_time: float = exported()

    def preprocess(self):
        if Options.mirror:
            self.lane = -self.lane
            self.direction = -self.direction

        self.judgment_window = note_judgment_window
        self.target_time = beat_to_time(self.beat)
        self.target_scaled_time = time_to_scaled_time(self.target_time)
        self.start_scaled_time = self.target_scaled_time - preempt_time()
        self.input_interval = self.judgment_window.good + self.target_time + input_offset()
        self.result.bucket = get_note_bucket(self.kind)
        self.result.accuracy = 1.0

        self.best_judgment_time = DEFAULT_BEST_JUDGMENT_TIME

        self.head_ref = self.ref()
        while self.head.has_prev:
            self.head_ref = self.head.prev_ref

        self.end_ref = self.ref()
        while self.end.has_next:
            self.end_ref = self.end.next_ref

        if Options.auto_sfx_enabled:
            schedule_note_sfx(self.kind, Judgment.PERFECT, self.target_time)

    def should_spawn(self) -> bool:
        return scaled_time() >= self.start_scaled_time

    def spawn_order(self) -> float:
        return self.start_scaled_time

    def initialize(self):
        if self.has_next and not self.has_prev:
            Streams.hold_activity[self.head.index][-10] = False
            HoldManager.spawn(head_ref=self.head_ref, end_ref=self.end_ref)

    def update_sequential(self):
        if self.kind == NoteKind.HOLD_ANCHOR:
            self.despawn = True
            return
        if time() > self.input_interval.end:
            self.despawn = True
            return
        if time() in self.input_interval:
            NoteMemory.active_notes.append(self.ref())
        if self.best_judgment_time > DEFAULT_BEST_JUDGMENT_TIME:
            # For holds ticks and flicks, we wait until it's impossible to improve the judgment before judging.
            # E.g. the player might be within a hold tick's hitbox at the early good window, move their finger away,
            # then move it back inside within the perfect window. It would be unfair to judge the note immediately
            # when the player moved their finger away, so we wait until it would be impossible to improve the judgment
            # before judging.
            can_improve = (
                self.best_judgment_time < self.target_time
                and offset_adjusted_time() - self.target_time < self.target_time - self.best_judgment_time
            )
            if not can_improve:
                if self.target_time - 1 / 30 <= self.best_judgment_time <= self.target_time:
                    # If the best judgment time is just before the target time, we assume that they player has been
                    # continuous holding the tick or flicking, and we treat it as having perfect timing.
                    self.best_judgment_time = self.target_time
                self.judge(self.best_judgment_time)
        # Compared to the original Bandori mechanics, pydori is more lenient with releasing hold notes.
        # Releasing a hold is no longer an automatic miss, so accidentally releasing a hold note early is okay
        # as long as the player puts their finger back down before the next tick or the hold end.
        # We still keep track of a single active touch so players can't cheat using multiple fingers on the same hold.
        if self.has_active_touch:
            claim_touch(self.active_touch_id)

    def update_parallel(self):
        if self.despawn:
            return
        draw_note(
            self.kind,
            self.lane,
            self.y,
            self.direction,
        )

    def touch(self):
        if self.despawn:
            return
        hitbox_quad = self.calculate_hitbox().layout()
        match self.kind:
            case NoteKind.TAP | NoteKind.HOLD_HEAD:
                self.handle_tap_input(hitbox_quad)
            case NoteKind.HOLD_TICK:
                self.handle_hold_input(hitbox_quad)
            case NoteKind.HOLD_END:
                self.handle_release_input(hitbox_quad)
            case NoteKind.FLICK | NoteKind.DIRECTIONAL_FLICK:
                self.handle_flick_input(hitbox_quad)

    def handle_tap_input(self, quad: Quad):
        if time() not in self.input_interval:
            return
        for touch in unclaimed_taps():
            if not quad.contains_point(touch.position):
                continue
            claim_touch(touch.id)
            self.active_touch_id = touch.id
            self.judge(touch.start_time)
            break

    def handle_hold_input(self, quad: Quad):
        if self.has_prev and not (self.head.is_judged or self.head.is_despawned):
            # If the hold head is still around, require players to tap the head to start the hold.
            return
        if time() not in self.input_interval:
            return
        self.capture_touch_if_needed(quad)
        if self.has_active_touch:
            for touch in touches():
                if touch.id != self.active_touch_id:
                    continue
                if quad.contains_point(touch.position):
                    self.update_best_judgment_time_with_current_time()
                if self.best_judgment_time >= self.target_time:
                    self.judge(self.best_judgment_time)
                break

    def handle_release_input(self, quad: Quad):
        if self.has_prev and not (self.head.is_judged or self.head.is_despawned):
            # If the hold head is still around, require players to tap the head to start the hold.
            return
        if time() not in self.input_interval:
            return
        self.capture_touch_if_needed(quad)
        if self.has_active_touch:
            for touch in touches():
                if touch.id != self.active_touch_id:
                    continue
                if touch.ended:
                    if quad.contains_point(touch.position):
                        self.judge(offset_adjusted_time())
                    else:
                        self.fail()
                break

    def handle_flick_input(self, quad: Quad):
        if self.has_prev and not (self.head.is_judged or self.head.is_despawned):
            # If the hold head is still around, require players to tap the head to start the hold.
            return
        if time() not in self.input_interval:
            return
        if self.has_prev:
            # If this is a hold-flick, then we don't require a new tap to start the flick.
            self.capture_touch_if_needed(quad)
        else:
            # Regular flicks require a tap before flicking.
            self.capture_tap_if_needed(quad)
        if self.has_active_touch:
            for touch in touches():
                if touch.id != self.active_touch_id:
                    continue
                meets_speed = touch.speed >= get_flick_speed_threshold(self.direction)
                meets_direction = self.direction == 0 or touch.velocity.dot(Vec2(self.direction, 0)) > 0
                if quad.contains_point(touch.position) and meets_speed and meets_direction:
                    self.update_best_judgment_time_with_current_time()
                if touch.ended or self.best_judgment_time >= self.target_time:
                    self.judge(self.best_judgment_time)
                break

    def capture_touch_if_needed(self, quad: Quad):
        """If there is no active touch, try to claim a touch that is inside the quad and set it as the active touch."""
        if not self.has_active_touch:
            for touch in unclaimed_touches():
                if not quad.contains_point(touch.position):
                    continue
                if touch.ended:
                    continue
                claim_touch(touch.id)
                self.active_touch_id = touch.id
                break

    def capture_tap_if_needed(self, quad: Quad):
        """If there is no active touch, try to claim a tap that is inside the quad and set it as the active touch."""
        if not self.has_active_touch:
            for touch in unclaimed_taps():
                if not quad.contains_point(touch.position):
                    continue
                claim_touch(touch.id)
                self.active_touch_id = touch.id
                break

    def update_best_judgment_time_with_current_time(self):
        prev_error = abs(self.best_judgment_time - self.target_time)
        new_error = abs(offset_adjusted_time() - self.target_time)
        if new_error < prev_error:
            self.best_judgment_time = offset_adjusted_time()

    def calculate_hitbox(self) -> Hitbox:
        base_hitbox = self.base_hitbox
        right_overlap = 0
        left_overlap = 0
        other_notes = (ref.get() for ref in NoteMemory.active_notes if ref.index != self.index)
        simultaneous_notes = (
            note
            for note in other_notes
            if not note.is_judged and abs(note.target_time - self.target_time) <= 0.005 and not note.has_active_touch
        )
        for sim_note in simultaneous_notes:
            sim_hitbox = sim_note.base_hitbox
            if sim_note.lane > self.lane:
                # The overlap between the hitboxes is how much the right side of the base hitbox
                # extends beyond the left side of the sim note's hitbox.
                right_overlap = max(right_overlap, base_hitbox.right - sim_hitbox.left)
            elif sim_note.lane < self.lane:
                # The same logic the other way around for the left side.
                left_overlap = max(left_overlap, sim_hitbox.right - base_hitbox.left)
        # Shrink the base hitbox by half the overlap on each side.
        return Hitbox(left=base_hitbox.left + left_overlap / 2, right=base_hitbox.right - right_overlap / 2)

    def terminate(self):
        self.end_time = time()

    def judge(self, judgment_time: float):
        judgment = self.judgment_window.judge(actual=judgment_time, target=self.target_time)
        self.result.judgment = judgment
        self.result.accuracy = clamp(judgment_time - self.target_time, -1.0, 1.0)
        self.result.bucket_value = self.result.accuracy * 1000
        if judgment != Judgment.MISS:
            if not Options.auto_sfx_enabled:
                play_note_sfx(self.kind, judgment)
            play_note_particle(self.kind, self.lane, self.direction)
        self.despawn = True
        self.is_judged = True

    def fail(self):
        self.result.judgment = Judgment.MISS
        self.result.accuracy = 1.0
        self.result.bucket_value = self.result.accuracy * 1000
        self.despawn = True
        self.is_judged = True

    @classmethod
    def global_preprocess(cls):
        init_note_life(cls)

    @property
    def kind(self) -> NoteKind:
        return cast(NoteKind, self.key)

    @property
    def y(self) -> float:
        return get_note_y(scaled_time(), self.target_scaled_time)

    @property
    def has_prev(self) -> bool:
        return self.prev_ref.index > 0

    @property
    def prev(self) -> Note:
        return self.prev_ref.get()

    @property
    def has_next(self) -> bool:
        return self.next_ref.index > 0

    @property
    def next(self) -> Note:
        return self.next_ref.get()

    @property
    def head(self) -> Note:
        return self.head_ref.get()

    @property
    def end(self) -> Note:
        return self.end_ref.get()

    @property
    def active_touch_id(self) -> int:
        return self.head._active_touch_id

    @active_touch_id.setter
    def active_touch_id(self, value: int):
        self.head._active_touch_id = value

    @property
    def has_active_touch(self) -> bool:
        return self.active_touch_id > 0

    @property
    def hold_lane(self) -> float:
        return self.head._hold_lane

    @hold_lane.setter
    def hold_lane(self, value: float):
        self.head._hold_lane = value

    @property
    def base_hitbox(self) -> Hitbox:
        return Hitbox.for_note(self.lane, self.direction)


TapNote = Note.derive("Tap", is_scored=True, key=NoteKind.TAP)
FlickNote = Note.derive("Flick", is_scored=True, key=NoteKind.FLICK)
DirectionalFlickNote = Note.derive("DirectionalFlick", is_scored=True, key=NoteKind.DIRECTIONAL_FLICK)
HoldHeadNote = Note.derive("HoldHead", is_scored=True, key=NoteKind.HOLD_HEAD)
HoldTickNote = Note.derive("HoldTick", is_scored=True, key=NoteKind.HOLD_TICK)
HoldAnchorNote = Note.derive("HoldAnchor", is_scored=False, key=NoteKind.HOLD_ANCHOR)
HoldEndNote = Note.derive("HoldEnd", is_scored=True, key=NoteKind.HOLD_END)

ALL_NOTE_TYPES = (
    TapNote,
    FlickNote,
    DirectionalFlickNote,
    HoldHeadNote,
    HoldTickNote,
    HoldAnchorNote,
    HoldEndNote,
)


class HoldManager(PlayArchetype):
    """Manages the particle and looping sfx of a hold note."""

    name = "HoldManager"

    head_ref: EntityRef[Note] = entity_memory()
    end_ref: EntityRef[Note] = entity_memory()

    particle: ParticleHandle = entity_memory()
    sfx: LoopedEffectHandle = entity_memory()

    def update_parallel(self):
        if time() > self.end.target_time:
            destroy_particle(self.particle)
            stop_looped_sfx(self.sfx)
            self.despawn = True
            return
        if time() < self.head.target_time:
            return
        Streams.hold_activity[self.head.index][time()] = self.head.has_active_touch
        if self.head.has_active_touch:
            draw_note_head(self.head.hold_lane)
            update_hold_particle(self.particle, self.head.hold_lane)
            update_hold_sfx(self.sfx)
        else:
            destroy_particle(self.particle)
            stop_looped_sfx(self.sfx)

    @callback(order=1)
    def touch(self):
        if not self.head.has_active_touch:
            return
        active_touch_id = self.head.active_touch_id
        for touch in touches():
            if touch.id != active_touch_id:
                continue
            if touch.ended:
                self.head.active_touch_id = 0
            break

    @property
    def head(self) -> Note:
        return self.head_ref.get()

    @property
    def end(self) -> Note:
        return self.end_ref.get()


@level_memory
class NoteMemory:
    active_notes: VarArray[EntityRef[Note], Dim[16]]



from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet
from sonolus.backend.optimize.flow import BasicBlock, traverse_cfg_preorder
from sonolus.backend.optimize.liveness import LivenessAnalysis, get_live
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig
from sonolus.backend.place import BlockPlace, SSAPlace, TempBlock


class CopyCoalesce(CompilerPass):
    def requires(self) -> set[CompilerPass]:
        return {LivenessAnalysis()}

    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        mapping = self.get_mapping(entry)
        for block in traverse_cfg_preorder(entry):
            block.statements = [self.apply_to_stmt(stmt, mapping) for stmt in block.statements]
            block.test = self.apply_to_stmt(block.test, mapping)
        return entry

    def apply_to_stmt(self, stmt, mapping: dict[TempBlock, TempBlock]):
        match stmt:
            case IRConst():
                return stmt
            case IRSet(place=place, value=value):
                return IRSet(self.apply_to_stmt(place, mapping), self.apply_to_stmt(value, mapping))
            case IRGet(place=place):
                return IRGet(self.apply_to_stmt(place, mapping))
            case IRPureInstr(op=op, args=args):
                return IRPureInstr(op, [self.apply_to_stmt(arg, mapping) for arg in args])
            case IRInstr(op=op, args=args):
                return IRInstr(op, [self.apply_to_stmt(arg, mapping) for arg in args])
            case BlockPlace(block=block, index=index, offset=offset):
                return BlockPlace(self.apply_to_stmt(block, mapping), self.apply_to_stmt(index, mapping), offset)
            case TempBlock():
                return mapping.get(stmt, stmt)
            case SSAPlace() | int() | float():
                return stmt

    def get_mapping(self, entry: BasicBlock) -> dict[TempBlock, TempBlock]:
        interference = self.get_interference(entry)
        copies = self.get_copies(entry)

        mapping: dict[TempBlock, set[TempBlock]] = {}

        copy_pairs = set()
        for target, sources in copies.items():
            copy_pairs.update((min(target, source), max(target, source)) for source in sources)

        for target, source in sorted(copy_pairs):
            if source in interference.get(target, set()):
                continue
            combined_mapping = mapping.get(target, {target}) | mapping.get(source, {source})
            combined_interference = interference.get(target, set()) | interference.get(source, set())
            for place in combined_mapping:
                mapping[place] = combined_mapping
                interference[place] = combined_interference
            for place in combined_interference:
                interference[place].update(combined_mapping)

        canonical_mapping = {}
        for place, group in mapping.items():
            if place in canonical_mapping:
                continue
            canonical = min(group)
            for member in group:
                canonical_mapping[member] = canonical

        return canonical_mapping

    def get_interference(self, entry: BasicBlock) -> dict[TempBlock, set[TempBlock]]:
        result = {}
        for block in traverse_cfg_preorder(entry):
            for stmt in block.statements:
                live = {p for p in get_live(stmt) if isinstance(p, TempBlock) and p.size == 1}
                for place in live:
                    result.setdefault(place, set()).update(live - {place})
            if not isinstance(block.test, IRConst):
                live = {p for p in get_live(block.test) if isinstance(p, TempBlock) and p.size == 1}
                for place in live:
                    result.setdefault(place, set()).update(live - {place})
        return result

    def get_copies(self, entry: BasicBlock) -> dict[TempBlock, set[TempBlock]]:
        result = {}
        for block in traverse_cfg_preorder(entry):
            for stmt in block.statements:
                if (
                    not isinstance(stmt, IRSet)
                    or not isinstance(stmt.place, BlockPlace)
                    or not isinstance(stmt.place.block, TempBlock)
                    or stmt.place.block.size != 1
                    or not isinstance(stmt.value, IRGet)
                    or not isinstance(stmt.value.place, BlockPlace)
                    or not isinstance(stmt.value.place.block, TempBlock)
                    or stmt.value.place.block.size != 1
                ):
                    continue
                target = stmt.place.block
                source = stmt.value.place.block
                result.setdefault(target, set()).add(source)
                result.setdefault(source, set()).add(target)
        return result



from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet
from sonolus.backend.ops import Op
from sonolus.backend.optimize.flow import BasicBlock, FlowEdge, traverse_cfg_preorder
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig
from sonolus.backend.place import BlockPlace, SSAPlace, TempBlock


class CoalesceFlow(CompilerPass):
    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        queue = [entry]
        processed = set()
        while queue:
            block = queue.pop()
            if block in processed:
                continue
            processed.add(block)
            for edge in block.outgoing:
                while True:
                    dst = edge.dst
                    if dst.phis or dst.statements or len(dst.outgoing) != 1 or dst is block or dst is entry:
                        break
                    next_dst = next(iter(dst.outgoing)).dst
                    if next_dst.phis:
                        break
                    dst.incoming.remove(edge)
                    if not dst.incoming:
                        for dst_edge in dst.outgoing:
                            dst_edge.dst.incoming.remove(dst_edge)
                        processed.add(dst)
                    edge.dst = next_dst
                    next_dst.incoming.add(edge)
                    if dst is edge.dst:
                        break
            default_edge = next((edge for edge in block.outgoing if edge.cond is None), None)
            if default_edge is not None:
                for edge in [*block.outgoing]:
                    if edge is default_edge:
                        continue
                    if edge.dst is default_edge.dst:
                        block.outgoing.remove(edge)
                        edge.dst.incoming.remove(edge)
            if len(block.outgoing) != 1:
                queue.extend(edge.dst for edge in block.outgoing)
                continue
            next_block = next(iter(block.outgoing)).dst
            if next_block is block or next_block is entry:
                continue
            if len(next_block.incoming) != 1:
                queue.append(next_block)
                if not block.statements and not block.phis and not next_block.phis:
                    for edge in block.incoming:
                        edge.dst = next_block
                        next_block.incoming.add(edge)
                    for edge in block.outgoing:  # There should be exactly one
                        next_block.incoming.remove(edge)
                    if block is entry:
                        entry = next_block
                continue
            for p, args in next_block.phis.items():
                if block not in args:
                    # This is the only predecessor to the block, so it must be a phi argument
                    raise ValueError("Missing phi argument")
                block.statements.append(IRSet(p, IRGet(args[block])))
            block.statements.extend(next_block.statements)
            block.test = next_block.test
            block.outgoing = next_block.outgoing
            for edge in block.outgoing:
                edge.src = block
                dst = edge.dst
                for args in dst.phis.values():
                    if next_block in args:
                        args[block] = args.pop(next_block)
            processed.add(next_block)
            queue.extend(edge.dst for edge in block.outgoing)
            processed.remove(block)
            queue.append(block)
        return entry


class CombineExitBlocks(CompilerPass):
    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        first_exit_block = None
        for block in traverse_cfg_preorder(entry):
            if not block.outgoing and not block.phis and not block.statements:
                if first_exit_block is None:
                    first_exit_block = block
                else:
                    for edge in [*block.incoming]:
                        edge.dst = first_exit_block
                        first_exit_block.incoming.add(edge)
                    block.incoming.clear()
        return entry


class CoalesceSmallConditionalBlocks(CompilerPass):
    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        queue = [entry]
        processed = set()
        while queue:
            block = queue.pop()
            if block.phis:
                raise RuntimeError("SSA form is not supported in this pass")
            if block in processed:
                continue
            processed.add(block)
            while len(block.outgoing) == 1:
                next_edge = next(iter(block.outgoing))
                next_block = next_edge.dst
                if len(next_block.statements) <= 1:
                    next_block.incoming.remove(next_edge)
                    block.test = next_block.test
                    block.outgoing = {FlowEdge(src=block, dst=edge.dst, cond=edge.cond) for edge in next_block.outgoing}
                    block.statements.extend(next_block.statements)
                    for edge in block.outgoing:
                        edge.dst.incoming.add(edge)
                else:
                    break
            queue.extend(
                edge.dst
                for edge in sorted(block.outgoing, key=lambda e: (e.cond is not None, e.cond))
                if edge.dst not in processed
            )

        reachable_blocks = set(traverse_cfg_preorder(entry))
        for block in traverse_cfg_preorder(entry):
            block.incoming = {edge for edge in block.incoming if edge.src in reachable_blocks}
        return entry


class RewriteToSwitch(CompilerPass):
    """Rewrite if-else chains to switch statements.

    Note that this needs inlining (and dead code elimination) to be run first to really do anything useful.
    """

    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        self.ifs_to_switch(entry)
        self.combine_blocks(entry)
        self.remove_unreachable(entry)
        return entry

    def ifs_to_switch(self, entry: BasicBlock):
        for block in traverse_cfg_preorder(entry):
            if len(block.outgoing) != 2 or {edge.cond for edge in block.outgoing} != {None, 0}:
                continue
            test = block.test
            if not isinstance(test, IRPureInstr) or test.op != Op.Equal:
                continue
            assert len(test.args) == 2
            if isinstance(test.args[0], IRConst):
                const, other = test.args
            elif isinstance(test.args[1], IRConst):
                other, const = test.args
            else:
                continue
            block.test = other
            for edge in block.outgoing:
                if edge.cond is None:
                    edge.cond = const.value
                else:
                    edge.cond = None

    def combine_blocks(self, entry: BasicBlock):
        queue = [entry]
        processed = set()
        while queue:
            block = queue.pop()
            if block in processed:
                continue
            processed.add(block)
            queue.extend(edge.dst for edge in block.outgoing)

            default_edge = next((edge for edge in block.outgoing if edge.cond is None), None)
            if default_edge is None:
                continue

            next_block = default_edge.dst
            if (
                len(next_block.incoming) > 1
                or next_block.statements
                or next_block.phis
                or block.test != next_block.test
                or block is next_block
                or next_block is entry
            ):
                continue

            outgoing_by_cond = {edge.cond: edge for edge in block.outgoing}
            assert len(outgoing_by_cond) == len(block.outgoing)
            outgoing_by_cond.pop(None)
            for edge in next_block.outgoing:
                if edge.cond in outgoing_by_cond:
                    # This edge is unreachable since an equivalent edge would have been taken
                    edge.dst.incoming.remove(edge)
                    continue
                outgoing_by_cond[edge.cond] = edge
                edge.src = block
                for args in edge.dst.phis.values():
                    if next_block in args:
                        args[block] = args.pop(next_block)
            block.outgoing = set(outgoing_by_cond.values())
            processed.add(next_block)
            queue.append(block)
            processed.remove(block)

    def remove_unreachable(self, entry: BasicBlock):
        reachable = {*traverse_cfg_preorder(entry)}
        for block in traverse_cfg_preorder(entry):
            block.incoming = {edge for edge in block.incoming if edge.src in reachable}
            block.outgoing = {edge for edge in block.outgoing if edge.dst in reachable}


class NormalizeSwitch(CompilerPass):
    """Normalize branches like cond -> case a, case a + b, case a + 2b to ((cond - a) / b) -> case 0, case 1, case 2."""

    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        for block in traverse_cfg_preorder(entry):
            cases = {edge.cond for edge in block.outgoing}
            if len(cases) <= 2:
                continue
            assert None in cases, "Non-terminal blocks should always have a default edge"
            cases.remove(None)
            offset, stride = self.get_offset_stride(cases)
            if offset is None or (offset == 0 and stride == 1):
                continue
            for edge in block.outgoing:
                if edge.cond is None:
                    continue
                edge.cond = (edge.cond - offset) // stride
            if offset != 0:
                block.test = IRPureInstr(Op.Subtract, [block.test, IRConst(offset)])
            if stride != 1:
                block.test = IRPureInstr(Op.Divide, [block.test, IRConst(stride)])
        return entry

    def get_offset_stride(self, cases: set[int]) -> tuple[int | None, int | None]:
        cases = sorted(cases)
        offset = cases[0]
        stride = cases[1] - offset
        if int(offset) != offset or int(stride) != stride:
            return None, None
        for i, case in enumerate(cases[2:], 2):
            if case != offset + i * stride:
                return None, None
        return offset, stride


class FlattenAssociativeOps(CompilerPass):
    FLATTENABLE_OPS = frozenset({Op.Add, Op.Multiply, Op.Mod, Op.Rem})

    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        for block in traverse_cfg_preorder(entry):
            block.statements = [self.flatten_statement(stmt) for stmt in block.statements]
            block.test = self.flatten_statement(block.test)
        return entry

    def flatten_statement(self, stmt):
        match stmt:
            case IRPureInstr(op=op, args=args) if op in self.FLATTENABLE_OPS:
                args = [self.flatten_statement(arg) for arg in args]
                if args and isinstance(args[0], IRPureInstr) and args[0].op == op:
                    args = [*args[0].args, *args[1:]]
                return IRPureInstr(op=op, args=args)
            case IRPureInstr(op=op, args=args):
                return IRPureInstr(op=op, args=[self.flatten_statement(arg) for arg in args])
            case IRInstr(op=op, args=args):
                return IRInstr(op=op, args=[self.flatten_statement(arg) for arg in args])
            case IRSet(place=place, value=value):
                return IRSet(place=place, value=self.flatten_statement(value))
            case _:
                return stmt


class RemoveRedundantArguments(CompilerPass):
    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        for block in traverse_cfg_preorder(entry):
            block.statements = [self.update_statement(stmt) for stmt in block.statements]
            block.test = self.update_statement(block.test)
        return entry

    def update_statement(self, stmt):
        match stmt:
            case IRPureInstr() | IRPureInstr():
                op = stmt.op
                args = stmt.args
                match op:
                    case Op.Add:
                        args = [arg for arg in args if not (isinstance(arg, IRConst) and arg.value == 0)]
                        if len(args) == 1:
                            return args[0]
                    case Op.Subtract:
                        args = [
                            args[0],
                            *(arg for arg in args[1:] if not (isinstance(arg, IRConst) and arg.value == 0)),
                        ]
                        if len(args) == 1:
                            return args[0]
                        if isinstance(args[0], IRConst) and args[0].value == 0 and len(args) == 2:
                            return IRPureInstr(Op.Negate, [self.update_statement(args[1])])
                    case Op.Multiply:
                        args = [arg for arg in args if not (isinstance(arg, IRConst) and arg.value == 1)]
                        if len(args) == 1:
                            return args[0]
                    case Op.Divide:
                        args = [
                            args[0],
                            *(arg for arg in args[1:] if not (isinstance(arg, IRConst) and arg.value == 1)),
                        ]
                        if len(args) == 1:
                            return args[0]
                return type(stmt)(op=op, args=[self.update_statement(arg) for arg in args])
            case IRSet(place=place, value=value):
                return IRSet(place=place, value=self.update_statement(value))
            case _:
                return stmt


class RenumberVars(CompilerPass):
    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        numbers = {}
        for block in traverse_cfg_preorder(entry):
            for stmt in block.statements:
                if (
                    isinstance(stmt, IRSet)
                    and isinstance(stmt.place, BlockPlace)
                    and isinstance(stmt.place.block, TempBlock)
                    and stmt.place.block.size == 1
                    and stmt.place.block not in numbers
                ):
                    numbers[stmt.place.block] = len(numbers) + 1
        for block in traverse_cfg_preorder(entry):
            block.statements = [self.update_statement(stmt, numbers) for stmt in block.statements]
            block.test = self.update_statement(block.test, numbers)
        return entry

    def update_statement(self, stmt, numbers: dict[TempBlock, int]):
        match stmt:
            case IRConst():
                return stmt
            case IRPureInstr(op=op, args=args):
                return IRPureInstr(op=op, args=[self.update_statement(arg, numbers) for arg in args])
            case IRInstr(op=op, args=args):
                return IRInstr(op=op, args=[self.update_statement(arg, numbers) for arg in args])
            case IRGet(place=place):
                return IRGet(place=self.update_statement(place, numbers))
            case IRSet(place=SSAPlace() as place, value=value):
                return IRSet(place=place, value=self.update_statement(value, numbers))
            case IRSet(place=place, value=value):
                return IRSet(place=self.update_statement(place, numbers), value=self.update_statement(value, numbers))
            case BlockPlace(block=block, index=index, offset=offset):
                return BlockPlace(
                    block=self.update_statement(block, numbers),
                    index=self.update_statement(index, numbers),
                    offset=offset,
                )
            case SSAPlace():
                return stmt
            case TempBlock() as b if b in numbers:
                return TempBlock(f"v{numbers[b]}", size=1)
            case int() | float() | TempBlock():
                return stmt
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")



from sonolus.build.engine import JsonValue, package_data
from sonolus.script.level import LevelData


def package_level_data(
    level_data: LevelData,
) -> bytes:
    return package_data(build_level_data(level_data))


def build_level_data(
    level_data: LevelData,
) -> JsonValue:
    level_refs = {entity: f"{i}_{entity.name}" for i, entity in enumerate(level_data.entities)}
    return {
        "bgmOffset": level_data.bgm_offset,
        "entities": [
            {
                "name": level_refs[entity],
                "archetype": entity.name,
                "data": entity._level_data_entries(level_refs),
            }
            for entity in level_data.entities
        ],
    }



from __future__ import annotations

import random
from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import Any, Final

from sonolus.script.debug import assert_true
from sonolus.script.internal.context import ctx
from sonolus.script.internal.math_impls import _trunc
from sonolus.script.internal.meta_fn import meta_fn
from sonolus.script.iterator import SonolusIterator
from sonolus.script.maybe import Maybe, Nothing, Some
from sonolus.script.num import Num
from sonolus.script.record import Record
from sonolus.script.values import copy

# Note: we don't use Range in this file because Range itself inherits from ArrayLike


class ArrayLike[T](Sequence[T]):
    """Mixin for array-like objects.

    Inheritors must implement `__len__`, `__getitem__`, and `__setitem__`.

    Usage:
        ```python
        class MyArrayLike[T](Record, ArrayLike[T]):
            def __len__(self) -> int:
                ...

            def __getitem__(self, index: int) -> T:
                ...

            def __setitem__(self, index: int, value: T):
                ...
        ```
    """

    _allow_instance_check_ = True

    @abstractmethod
    def __len__(self) -> int:
        """Return the length of the array."""

    @abstractmethod
    def __getitem__(self, index: int) -> T:
        """Return the item at the given index.

        Args:
            index: The index of the item. Must be an integer between 0 and `len(self) - 1`.
        """

    @abstractmethod
    def __setitem__(self, index: int, value: T):
        """Set the value of the item at the given index.

        Args:
            index: The index of the item. Must be an integer between 0 and `len(self) - 1`.
            value: The value to set.
        """

    @meta_fn
    def get_unchecked(self, index: Num) -> T:
        """Get the element at the given index possibly without bounds checking or conversion of negative indexes.

        The compiler may still determine that the index is out of bounds and throw an error, but it may skip these
        checks at runtime.

        Args:
            index: The index to get.

        Returns:
            The element at the given index.
        """
        return self[index]

    @meta_fn
    def set_unchecked(self, index: Num, value: T):
        """Set the element at the given index possibly without bounds checking or conversion of negative indexes.

        The compiler may still determine that the index is out of bounds and throw an error, but it may skip these
        checks at runtime.

        Args:
            index: The index to set.
            value: The value to set.
        """
        self[index] = value

    def __iter__(self) -> SonolusIterator[T]:
        """Return an iterator over the array."""
        return _ArrayIterator(0, self.unchecked())

    def __contains__(self, value: Any) -> bool:
        """Return whether any element in the array is equal to the given value.

        Args:
            value: The value to check for.
        """
        i = 0
        while i < len(self):
            if self.get_unchecked(i) == value:
                return True
            i += 1
        return False

    def __reversed__(self):
        """Return a reversed view of the array."""
        return _ArrayReverser(self)

    def _enumerate_(self, start: int = 0) -> SonolusIterator[T]:
        return _ArrayEnumerator(0, start, self)

    def index(self, value: T, start: int = 0, stop: int | None = None) -> int:
        """Return the index of the value in the array equal to the given value.

        Args:
            value: The value to search for.
            start: The index to start searching from.
            stop: The index to stop searching at. If `None`, search to the end of the array.
        """
        if stop is None:
            stop = len(self)
        else:
            stop = get_positive_index(stop, len(self), check=False)
        stop = min(stop, len(self))
        start = get_positive_index(start, len(self), check=False)
        i = max(start, 0)
        while i < stop:
            if self.get_unchecked(i) == value:
                return i
            i += 1
        return -1

    def count(self, value: T) -> int:
        """Return the number of elements in the array equal to the given value.

        Args:
            value: The value to count.
        """
        count = 0
        i = 0
        while i < len(self):
            if self.get_unchecked(i) == value:
                count += 1
            i += 1
        return count

    def last_index(self, value: T) -> int:
        """Return the last index of the value in the array equal to the given value.

        Args:
            value: The value to search for.
        """
        i = len(self) - 1
        while i >= 0:
            if self.get_unchecked(i) == value:
                return i
            i -= 1
        return -1

    def index_of_max(self, *, key: Callable[[T], Any] | None = None) -> int:
        """Return the index of the maximum value in the array.

        Args:
            key: A one-argument ordering function to use for comparison like the one used in `max()`.
        """
        if len(self) == 0:
            return -1
        if key is None:
            key = _identity  # type: ignore
        max_index = 0
        i = 1
        while i < len(self):
            if key(self.get_unchecked(i)) > key(self.get_unchecked(max_index)):  # type: ignore
                max_index = i
            i += 1
        return max_index

    def index_of_min(self, *, key: Callable[[T], Any] | None = None) -> int:
        """Return the index of the minimum value in the array.

        Args:
            key: A one-argument ordering function to use for comparison like the one used in `min()`.
        """
        if len(self) == 0:
            return -1
        if key is None:
            key = _identity  # type: ignore
        min_index = 0
        i = 1
        while i < len(self):
            if key(self.get_unchecked(i)) < key(self.get_unchecked(min_index)):  # type: ignore
                min_index = i
            i += 1
        return min_index

    def _max_(self, key: Callable[[T], Any] | None = None) -> T:
        index = self.index_of_max(key=key)
        assert index != -1
        return self.get_unchecked(index)

    def _min_(self, key: Callable[[T], Any] | None = None) -> T:
        index = self.index_of_min(key=key)
        assert index != -1
        return self.get_unchecked(index)

    def swap(self, i: int, j: int, /):
        """Swap the values at the given positive indices.

        Args:
            i: The first index.
            j: The second index.
        """
        check_positive_index(i, len(self))
        check_positive_index(j, len(self))
        self.swap_unchecked(i, j)

    def swap_unchecked(self, i: Num, j: Num):
        """Swap the values at the given indices possibly without bounds checking.

        Args:
            i: The first index.
            j: The second index.
        """
        temp = copy(self.get_unchecked(i))
        self.set_unchecked(i, self.get_unchecked(j))
        self.set_unchecked(j, temp)

    def sort(self, *, key: Callable[[T], Any] | None = None, reverse: bool = False):
        """Sort the values in the array in place.

        Args:
            key: A one-argument ordering function to use for comparison.
            reverse: If `True`, sort in descending order, otherwise sort in ascending order.
        """
        if key is not None or len(self) < 15:
            if key is None:
                key = _identity  # type: ignore
            # May be worth adding a block sort variant for better performance on large arrays in the future
            _insertion_sort(self.unchecked(), 0, len(self), key, reverse)  # type: ignore
        else:
            # Heap sort is unstable, so if there's a key, we can't rely on it
            _heap_sort(self.unchecked(), 0, len(self), reverse)  # type: ignore

    def shuffle(self):
        """Shuffle the values in the array in place."""
        random.shuffle(self)  # type: ignore

    def reverse(self):
        """Reverse the values in the array in place."""
        i = 0
        j = len(self) - 1
        while i < j:
            self.swap_unchecked(i, j)
            i += 1
            j -= 1

    def unchecked(self) -> ArrayLike[T]:
        """Return a proxy object that may skip bounds checking and may not support negative indexes."""
        return UncheckedArrayProxy(self)


def _identity[T](value: T) -> T:
    return value


def _insertion_sort[T](array: ArrayLike[T], start: int, end: int, key: Callable[[T], Any], reverse: bool):
    i = start + 1
    if reverse:
        while i < end:
            j = i
            while j > start and key(array[j - 1]) < key(array[j]):  # type: ignore
                array.swap_unchecked(j - 1, j)
                j -= 1
            i += 1
    else:
        while i < end:
            j = i
            while j > start and key(array[j - 1]) > key(array[j]):  # type: ignore
                array.swap_unchecked(j - 1, j)
                j -= 1
            i += 1


def _heapify[T](array: ArrayLike[T], end: int, index: int, reverse: bool):
    while True:
        left = index * 2 + 1
        right = left + 1
        largest = index
        if left < end and (array[left] > array[largest]) != reverse:  # type: ignore
            largest = left
        if right < end and (array[right] > array[largest]) != reverse:  # type: ignore
            largest = right
        if largest == index:
            break
        array.swap_unchecked(index, largest)
        index = largest


def _heap_sort[T](array: ArrayLike[T], start: int, end: int, reverse: bool):
    i = end // 2 - 1
    while i >= start:
        _heapify(array, end, i, reverse)
        i -= 1
    i = end - 1
    while i > start:
        array.swap_unchecked(start, i)
        _heapify(array, i, start, reverse)
        i -= 1


class _ArrayIterator[V: ArrayLike](Record, SonolusIterator):
    i: int
    array: V

    def next(self) -> Maybe[V]:
        if self.i < len(self.array):
            value = self.array.get_unchecked(self.i)
            self.i += 1
            return Some(value)
        return Nothing


class _ArrayReverser[V: ArrayLike](Record, ArrayLike):
    array: V

    def __len__(self) -> int:
        return len(self.array)

    def __getitem__(self, index: int) -> V:
        return self.array[len(self) - 1 - index]

    def __setitem__(self, index: int, value: V):
        self.array[len(self) - 1 - index] = value

    def get_unchecked(self, index: Num) -> V:
        return self.array.get_unchecked(len(self) - 1 - index)

    def set_unchecked(self, index: Num, value: V):
        self.array.set_unchecked(len(self) - 1 - index, value)

    def reversed(self) -> ArrayLike[V]:
        return self.array


class _ArrayEnumerator[V: ArrayLike](Record, SonolusIterator):
    i: int
    offset: Final[int]
    array: V

    def next(self) -> Maybe[tuple[int, Any]]:
        if self.i < len(self.array):
            result = (self.i + self.offset, self.array.get_unchecked(self.i))
            self.i += 1
            return Some(result)
        return Nothing


@meta_fn
def get_positive_index(
    index: int | float, length: int | float, *, include_end: bool = False, check: bool = True
) -> int:
    """Get the positive index for the given index in the array of the given length, and also perform bounds checking.

    This is used to convert negative indices relative to the end of the array to positive indices.

    Args:
        index: The index to convert.
        length: The length of the array.
        include_end: Whether to allow the index to be equal to the length of the array (i.e., one past the end).
        check: Whether to perform bounds checking. Must be a compile-time constant.

    Returns:
        The positive integer index.
    """
    if not ctx():
        if check:
            if (include_end and not -length <= index <= length) or (not include_end and not -length <= index < length):
                raise IndexError("Index out of range")
            if int(index) != index:
                raise ValueError("Index must be an integer")
            if int(length) != length:
                raise ValueError("Length must be an integer")
            if length < 0:
                raise ValueError("Length must be non-negative")
        return int(index + (index < 0) * length)
    index = Num._accept_(index)
    length = Num._accept_(length)
    if Num._accept_(check)._as_py_():
        include_end = Num._accept_(include_end)
        if not include_end._is_py_():
            is_in_bounds = Num.and_(index >= -length, index < (length + include_end))
        elif include_end._as_py_():
            is_in_bounds = Num.and_(index >= -length, index <= length)
        else:
            is_in_bounds = Num.and_(index >= -length, index < length)
        assert_true(Num.and_(is_in_bounds, _trunc(index) == index), "Invalid index")
        # Skipping length check since typically these are managed by the library and unlikely to be wrong
    return index + (index < 0) * length


@meta_fn
def check_positive_index(index: int, length: int, include_end: bool = False) -> int | float:
    """Check that the given index is a valid index for the array of the given length and convert it to an integer.

    Args:
        index: The index to check.
        length: The length of the array.
        include_end: Whether to allow the index to be equal to the length of the array (i.e., one past the end).

    Returns:
        The index as an integer.
    """
    if not ctx():
        if (include_end and not 0 <= index <= length) or (not include_end and not 0 <= index < length):
            raise IndexError("Index out of range")
        if int(index) != index:
            raise ValueError("Index must be an integer")
        if int(length) != length:
            raise ValueError("Length must be an integer")
        if length < 0:
            raise ValueError("Length must be non-negative")
        return int(index)
    index = Num._accept_(index)
    length = Num._accept_(length)
    include_end = Num._accept_(include_end)
    if not include_end._is_py_():
        is_in_bounds = Num.and_(index >= 0, index < (length + include_end))
    elif include_end._as_py_():
        is_in_bounds = Num.and_(index >= 0, index <= length)
    else:
        is_in_bounds = Num.and_(index >= 0, index < length)
    assert_true(Num.and_(is_in_bounds, _trunc(index) == index), "Invalid index")
    # Skipping length check since typically these are managed by the library and unlikely to be wrong
    return index


class UncheckedArrayProxy[T](Record, ArrayLike):
    array: T

    def __len__(self) -> int:
        return len(self.array)

    def __getitem__(self, index: int) -> Any:
        return self.array.get_unchecked(index)

    def __setitem__(self, index: int, value: Any):
        self.array.set_unchecked(index, value)



from __future__ import annotations

from sonolus.backend.ops import Op
from sonolus.script.array_like import ArrayLike
from sonolus.script.debug import static_error
from sonolus.script.internal.builtin_impls import _max, _min
from sonolus.script.internal.meta_fn import perf_meta_fn
from sonolus.script.internal.native import native_function
from sonolus.script.internal.range import range_or_tuple
from sonolus.script.num import Num
from sonolus.script.record import Record


class Interval(Record):
    """A closed interval.

    Usage:
        ```python
        Interval(start: float, end: float)
        ```
    """

    start: float
    end: float

    @classmethod
    def zero(cls) -> Interval:
        """Get an empty interval."""
        return cls(0, 0)

    @property
    def length(self) -> float:
        """The length of the interval.

        May be negative if the end is less than the start.
        """
        return self.end - self.start

    @property
    def is_empty(self) -> bool:
        """Whether the interval has a start greater than its end."""
        return self.start > self.end

    @property
    def mid(self) -> float:
        """The midpoint of the interval."""
        return (self.start + self.end) / 2

    @property
    def tuple(self):
        """The interval as a tuple."""
        return self.start, self.end

    def __contains__(self, item: Interval | float | int) -> bool:
        """Check if an item is within the interval.

        Args:
            item: The item to check. If it is an interval, it must be fully contained within this interval.

        Returns:
            True if the item is within the interval, False otherwise.
        """
        match item:
            case Interval(start, end):
                return self.start <= start and end <= self.end
            case Num(value):
                return self.start <= value <= self.end
            case _:
                static_error("Invalid type for interval check")

    @perf_meta_fn
    def __add__(self, other: float | int) -> Interval:
        """Add a value to both ends of the interval.

        Args:
            other: The value to add.

        Returns:
            A new interval with the value added to both ends.
        """
        return Interval._unchecked(start=self.start + other, end=self.end + other)

    @perf_meta_fn
    def __sub__(self, other: float | int) -> Interval:
        """Subtract a value from both ends of the interval.

        Args:
            other: The value to subtract.

        Returns:
            A new interval with the value subtracted from both ends.
        """
        return Interval._unchecked(start=self.start - other, end=self.end - other)

    @perf_meta_fn
    def __mul__(self, other: float | int) -> Interval:
        """Multiply both ends of the interval by a value.

        Args:
            other: The value to multiply by.

        Returns:
            A new interval with both ends multiplied by the value.
        """
        return Interval._unchecked(start=self.start * other, end=self.end * other)

    @perf_meta_fn
    def __truediv__(self, other: float | int) -> Interval:
        """Divide both ends of the interval by a value.

        Args:
            other: The value to divide by.

        Returns:
            A new interval with both ends divided by the value.
        """
        return Interval._unchecked(start=self.start / other, end=self.end / other)

    @perf_meta_fn
    def __floordiv__(self, other: float | int) -> Interval:
        """Divide both ends of the interval by a value and floor the result.

        Args:
            other: The value to divide by.

        Returns:
            A new interval with both ends divided by the value and floored.
        """
        return Interval._unchecked(start=self.start // other, end=self.end // other)

    def __and__(self, other: Interval) -> Interval:
        """Get the intersection of two intervals.

        The resulting interval will be empty and may have a negative length if the two intervals do not overlap.

        Args:
            other: The other interval.

        Returns:
            A new interval representing the intersection of the two intervals.
        """
        return Interval(max(self.start, other.start), min(self.end, other.end))

    def shrink(self, value: float | int) -> Interval:
        """Shrink the interval by a value on both ends.

        Args:
            value: The value to shrink by.

        Returns:
            A new interval with the value subtracted from the start and added to the end.
        """
        return Interval(self.start + value, self.end - value)

    def expand(self, value: float | int) -> Interval:
        """Expand the interval by a value on both ends.

        Args:
            value: The value to expand by.

        Returns:
            A new interval with the value subtracted from the start and added to the end.
        """
        return Interval(self.start - value, self.end + value)

    def lerp(self, x: float, /) -> float:
        """Linearly interpolate a value within the interval.

        Args:
            x: The interpolation factor.

        Returns:
            The interpolated value.
        """
        return lerp(self.start, self.end, x)

    def lerp_clamped(self, x: float, /) -> float:
        """Linearly interpolate a value within the interval, clamped to the interval.

        Args:
            x: The interpolation factor.

        Returns:
            The interpolated value.
        """
        return lerp_clamped(self.start, self.end, x)

    def unlerp(self, x: float, /) -> float:
        """Inverse linear interpolation of a value within the interval.

        Args:
            x: The value to unlerp.

        Returns:
            The unlerped value.
        """
        return unlerp(self.start, self.end, x)

    def unlerp_clamped(self, x: float, /) -> float:
        """Inverse linear interpolation of a value within the interval, clamped to the interval.

        Args:
            x: The value to unlerp.

        Returns:
            The unlerped value.
        """
        return unlerp_clamped(self.start, self.end, x)

    def clamp(self, x: float, /) -> float:
        """Clamp a value to the interval.

        Args:
            x: The value to clamp.

        Returns:
            The clamped value.
        """
        return clamp(x, self.start, self.end)


@native_function(Op.Lerp)
def _num_lerp(a, b, x, /):
    return a + (b - a) * x


@native_function(Op.LerpClamped)
def _num_lerp_clamped(a, b, x, /):
    return a + (b - a) * max(0, min(1, x))


@perf_meta_fn
def _generic_lerp[T](a: T, b: T, x: float, /) -> T:
    return a + (b - a) * x  # type: ignore


@perf_meta_fn
def _generic_lerp_clamped[T](a: T, b: T, x: float, /) -> T:
    return a + (b - a) * _max(0, _min(1, x))  # type: ignore


@perf_meta_fn
def lerp[T](a: T, b: T, x: float, /) -> T:
    """Linearly interpolate between two values.

    Args:
        a: The start value.
        b: The end value.
        x: The interpolation factor.

    Returns:
        The interpolated value.
    """
    match a, b:
        case (Num(a), Num(b)):
            return _num_lerp(a, b, x)
        case _:
            return _generic_lerp(a, b, x)


@perf_meta_fn
def lerp_clamped[T](a: T, b: T, x: float, /) -> T:
    """Linearly interpolate between two values, clamped to the interval.

    Args:
        a: The start value.
        b: The end value.
        x: The interpolation factor.

    Returns:
        The interpolated value.
    """
    match a, b:
        case (Num(a), Num(b)):
            return _num_lerp_clamped(a, b, x)
        case _:
            return _generic_lerp_clamped(a, b, x)


@native_function(Op.Unlerp)
def unlerp(a: float, b: float, x: float, /) -> float:
    """Inverse linear interpolation.

    Args:
        a: The start value.
        b: The end value.
        x: The value to unlerp.

    Returns:
        The unlerped value.
    """
    return (x - a) / (b - a)


@native_function(Op.UnlerpClamped)
def unlerp_clamped(a: float, b: float, x: float, /) -> float:
    """Inverse linear interpolation, clamped to the interval.

    Args:
        a: The start value.
        b: The end value.
        x: The value to unlerp.

    Returns:
        The unlerped value.
    """
    return max(0, min(1, (x - a) / (b - a)))


@native_function(Op.Remap)
def remap(a: float, b: float, c: float, d: float, x: float, /) -> float:
    """Linearly remap a value from one interval to another.

    Args:
        a: The start of the input interval.
        b: The end of the input interval.
        c: The start of the output interval.
        d: The end of the output interval.
        x: The value to remap.

    Returns:
        The remapped value.
    """
    return c + (d - c) * (x - a) / (b - a)


@native_function(Op.RemapClamped)
def remap_clamped(a: float, b: float, c: float, d: float, x: float, /) -> float:
    """Linearly remap a value from one interval to another, clamped to the output interval.

    Args:
        a: The start of the input interval.
        b: The end of the input interval.
        c: The start of the output interval.
        d: The end of the output interval.
        x: The value to remap.

    Returns:
        The remapped value.
    """
    return c + (d - c) * max(0, min(1, (x - a) / (b - a)))


@native_function(Op.Clamp, const_eval=True)
def clamp(x: float, a: float, b: float, /) -> float:
    """Clamp a value to an interval.

    Args:
        x: The value to clamp.
        a: The start of the interval.
        b: The end of the interval.

    Returns:
        The clamped value.
    """
    return max(a, min(b, x))


def interp(
    xp: ArrayLike[float] | tuple[float, ...],
    fp: ArrayLike[float] | tuple[float, ...],
    x: float,
) -> float:
    """Linearly interpolate a value within a sequence of points.

    The sequence must have at least 2 elements and be sorted in increasing order of x-coordinates.
    For values of x outside the range of xp, the slope of the first or last segment is used to extrapolate.

    Args:
        xp: The x-coordinates of the points in increasing order.
        fp: The y-coordinates of the points.
        x: The x-coordinate to interpolate.

    Returns:
        The interpolated value.
    """
    assert len(xp) == len(fp), "xp and fp must have the same length"
    assert len(xp) >= 2, "xp and fp must have at least 2 elements"
    for i in range_or_tuple(1, len(xp) - 1):
        assert xp[i] > xp[i - 1], "xp must be in increasing order"
        # At i == 1, x may be less than x[0], but since we're extrapolating, we use the first segment regardless.
        if x <= xp[i]:
            return remap(xp[i - 1], xp[i], fp[i - 1], fp[i], x)
    # x > xp[-2] so we can just use the last segment regardless of whether x is in it or to the right of it.
    return remap(xp[-2], xp[-1], fp[-2], fp[-1], x)


def interp_clamped(
    xp: ArrayLike[float] | tuple[float, ...],
    fp: ArrayLike[float] | tuple[float, ...],
    x: float,
) -> float:
    """Linearly interpolate a value within a sequence of points.

    The sequence must have at least 2 elements and be sorted in increasing order of x-coordinates.
    For x-coordinates outside the range of the sequence, the respective endpoint of fp is returned.

    Args:
        xp: The x-coordinates of the points in increasing order.
        fp: The y-coordinates of the points.
        x: The x-coordinate to interpolate.

    Returns:
        The interpolated value.
    """
    assert len(xp) == len(fp), "xp and fp must have the same length"
    assert len(xp) >= 2, "xp and fp must have at least 2 elements"
    if x <= xp[0]:
        return fp[0]
    for i in range_or_tuple(1, len(xp)):
        assert xp[i] > xp[i - 1], "xp must be in increasing order"
        if x <= xp[i]:
            return remap(xp[i - 1], xp[i], fp[i - 1], fp[i], x)
    return fp[-1]



from sonolus.backend.ops import Op
from sonolus.backend.place import BlockPlace, Place, SSAPlace

type IRExpr = IRConst | IRPureInstr | IRGet
type IRStmt = IRExpr | IRInstr | IRSet


_IR_CONST_CACHE_START = -5
_IR_CONST_CACHE_STOP = 257


class IRConst:
    __slots__ = ("value",)

    value: float | int

    def __new__(cls, value):
        if float(value).is_integer():
            int_value = int(value)
            if _IR_CONST_CACHE_START <= int_value < _IR_CONST_CACHE_STOP:
                return _IR_CONST_CACHE[int_value - _IR_CONST_CACHE_START]
            else:
                return _create_raw_const(int_value)
        return super().__new__(cls)

    def __init__(self, value: float):
        self.value = value

    def __repr__(self):
        return f"IRConst({self.value!r})"

    def __str__(self):
        return f"{self.value}"

    def __eq__(self, other):
        return isinstance(other, IRConst) and self.value == other.value

    def __hash__(self):
        return hash(self.value)


def _create_raw_const(value: float | int) -> IRConst:
    result = object.__new__(IRConst)
    result.value = value
    return result


_IR_CONST_CACHE = tuple(_create_raw_const(i) for i in range(_IR_CONST_CACHE_START, _IR_CONST_CACHE_STOP))


def format_ir(instr, parenthesize: bool = True) -> str:
    if not isinstance(instr, IRPureInstr | IRInstr):
        return str(instr)

    def apply_parentheses(s: str) -> str:
        return f"({s})" if parenthesize else s

    match instr.op:
        case Op.Add:
            return apply_parentheses(" + ".join(format_ir(arg, True) for arg in instr.args))
        case Op.Subtract:
            return apply_parentheses(" - ".join(format_ir(arg, True) for arg in instr.args))
        case Op.Multiply:
            return apply_parentheses(" * ".join(format_ir(arg, True) for arg in instr.args))
        case Op.Divide:
            return apply_parentheses(" / ".join(format_ir(arg, True) for arg in instr.args))
        case Op.Negate:
            return f"-{format_ir(instr.args[0], True)}"
        case Op.Mod:
            return apply_parentheses(" % ".join(format_ir(arg, True) for arg in instr.args))
        case Op.And:
            return apply_parentheses(" && ".join(format_ir(arg, True) for arg in instr.args))
        case Op.Or:
            return apply_parentheses(" || ".join(format_ir(arg, True) for arg in instr.args))
        case Op.Not:
            return f"!{format_ir(instr.args[0], True)}"
        case Op.Equal:
            return apply_parentheses(" == ".join(format_ir(arg, True) for arg in instr.args))
        case Op.NotEqual:
            return apply_parentheses(" != ".join(format_ir(arg, True) for arg in instr.args))
        case Op.Less:
            return apply_parentheses(" < ".join(format_ir(arg, True) for arg in instr.args))
        case Op.LessOr:
            return apply_parentheses(" <= ".join(format_ir(arg, True) for arg in instr.args))
        case Op.Greater:
            return apply_parentheses(" > ".join(format_ir(arg, True) for arg in instr.args))
        case Op.GreaterOr:
            return apply_parentheses(" >= ".join(format_ir(arg, True) for arg in instr.args))
        case _:
            return f"{instr.op.name}(" + ", ".join(format_ir(arg, False) for arg in instr.args) + ")"


class IRPureInstr:
    __slots__ = ("args", "array_defs", "defs", "is_array_init", "live", "op", "uses", "visited")

    op: Op
    args: list[IRExpr]

    def __init__(self, op: Op, args: list[IRExpr]):
        assert op.pure
        self.op = op
        self.args = args

    def __repr__(self):
        return f"IRPureInstr({self.op!r}, {self.args!r})"

    def __str__(self):
        return format_ir(self)

    def __eq__(self, other):
        return isinstance(other, IRPureInstr) and self.op == other.op and self.args == other.args

    def __hash__(self):
        return hash((self.op, tuple(self.args)))


class IRInstr:
    __slots__ = ("args", "array_defs", "defs", "is_array_init", "live", "op", "uses", "visited")

    op: Op
    args: list[IRExpr]

    def __init__(self, op: Op, args: list[IRExpr]):
        self.op = op
        self.args = args

    def __repr__(self):
        return f"IRInstr({self.op!r}, {self.args!r})"

    def __str__(self):
        return format_ir(self)

    def __eq__(self, other):
        return isinstance(other, IRInstr) and self.op == other.op and self.args == other.args

    def __hash__(self):
        return hash((self.op, tuple(self.args)))


class IRGet:
    __slots__ = ("array_defs", "defs", "is_array_init", "live", "place", "uses", "visited")

    place: Place

    def __init__(self, place: Place):
        self.place = place

    def __repr__(self):
        return f"IRGet({self.place!r})"

    def __str__(self):
        return f"{self.place}"

    def __eq__(self, other):
        return isinstance(other, IRGet) and self.place == other.place

    def __hash__(self):
        return hash(self.place)


class IRSet:
    __slots__ = ("array_defs", "defs", "is_array_init", "live", "place", "uses", "value", "visited")

    place: Place
    value: IRExpr | IRInstr

    def __init__(self, place: Place, value: IRExpr | IRInstr):
        self.place = place
        self.value = value

    def __repr__(self):
        return f"IRSet({self.place!r}, {self.value!r})"

    def __str__(self):
        match self.place:
            case BlockPlace():
                return f"{self.place} <- {format_ir(self.value, False)}"
            case SSAPlace():
                return f"{self.place} := {format_ir(self.value, False)}"
            case _:
                raise TypeError(f"Invalid place: {self.place}")

    def __eq__(self, other):
        return isinstance(other, IRSet) and self.place == other.place and self.value == other.value

    def __hash__(self):
        return hash((self.place, self.value))



# ruff: noqa: PLR1702
import functools
import math
import operator
from collections import defaultdict
from typing import ClassVar

import sonolus.script.internal.math_impls as smath
from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet, IRStmt
from sonolus.backend.ops import Op
from sonolus.backend.optimize.flow import BasicBlock, FlowEdge, traverse_cfg_preorder
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig
from sonolus.backend.place import BlockPlace, SSAPlace, TempBlock


class Undefined:
    pass


class NotAConstant:
    pass


UNDEF = Undefined()
NAC = NotAConstant()


type Value = float | set[float] | Undefined | NotAConstant


def is_constant(value: Value) -> bool:
    return isinstance(value, (int, float))


class SparseConditionalConstantPropagation(CompilerPass):
    SUPPORTED_OPS: ClassVar[set[Op]] = {
        Op.Equal,
        Op.NotEqual,
        Op.Greater,
        Op.GreaterOr,
        Op.Less,
        Op.LessOr,
        Op.Not,
        Op.And,
        Op.Or,
        Op.Negate,
        Op.Add,
        Op.Subtract,
        Op.Multiply,
        Op.Divide,
        Op.Power,
        Op.Log,
        Op.Ceil,
        Op.Floor,
        Op.Round,
        Op.Frac,
        Op.Mod,
        Op.Rem,
        Op.Sin,
        Op.Cos,
        Op.Tan,
        Op.Sinh,
        Op.Cosh,
        Op.Tanh,
        Op.Arcsin,
        Op.Arccos,
        Op.Arctan,
        Op.Arctan2,
        Op.Max,
        Op.Min,
        Op.Abs,
        Op.Clamp,
        Op.Degree,
        Op.Radian,
        Op.Lerp,
        Op.LerpClamped,
        Op.Remap,
        Op.RemapClamped,
    }

    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        ssa_edges: dict[SSAPlace, set[SSAPlace | BasicBlock]] = {}
        executable_edges: set[FlowEdge] = set()

        # BasicBlock key means the block's test
        values: dict[SSAPlace | BasicBlock, Value] = defaultdict(lambda: UNDEF)
        defs: dict[SSAPlace | BasicBlock, IRStmt | dict[FlowEdge, SSAPlace]] = {}
        places_to_blocks: dict[SSAPlace, BasicBlock] = {}
        reachable_blocks: set[BasicBlock] = set()

        for block in traverse_cfg_preorder(entry):
            incoming_by_src = {}
            for edge in block.incoming:
                incoming_by_src.setdefault(edge.src, []).append(edge)
            for p, args in block.phis.items():
                if not isinstance(p, SSAPlace):
                    raise TypeError(f"Unexpected phi place: {p}")
                defs[p] = {}
                for b, v in args.items():
                    for incoming in incoming_by_src.get(b, []):
                        defs[p][incoming] = v
                values[p] = UNDEF
                for arg in args.values():
                    ssa_edges.setdefault(arg, set()).add(p)
            for stmt in block.statements:
                if isinstance(stmt, IRSet) and isinstance(stmt.place, SSAPlace):
                    defs[stmt.place] = stmt.value
                    places_to_blocks[stmt.place] = block
                    values[stmt.place] = UNDEF
                    for dep in self.get_dependencies(stmt.value, set()):
                        ssa_edges.setdefault(dep, set()).add(stmt.place)
            defs[block] = block.test
            values[block] = UNDEF
            for dep in self.get_dependencies(block.test, set()):
                ssa_edges.setdefault(dep, set()).add(block)

        def visit_phi(p):
            arg_values = [values[v] if e in executable_edges else UNDEF for e, v in defs[p].items()]
            distinct_defined_arg_values = {arg for arg in arg_values if arg is not UNDEF}
            value = values[p]
            if len(distinct_defined_arg_values) == 1:
                new_value = distinct_defined_arg_values.pop()
            elif len(distinct_defined_arg_values) > 1:
                if any(arg is NAC for arg in distinct_defined_arg_values):
                    new_value = NAC
                else:
                    new_values = set()
                    for arg in distinct_defined_arg_values:
                        if isinstance(arg, frozenset):
                            new_values.update(arg)
                        else:
                            new_values.add(arg)
                    if len(new_values) == 1:
                        new_value = next(iter(new_values))
                    elif len(new_values) > 100:
                        new_value = NAC  # Give up to save performance
                    else:
                        new_value = frozenset(new_values)
            else:
                new_value = UNDEF
            if new_value != value:
                values[p] = new_value
                ssa_worklist.update(ssa_edges.get(p, set()))

        flow_worklist: set[FlowEdge] = {FlowEdge(entry, entry, None)}
        ssa_worklist: set[SSAPlace | BasicBlock] = set()
        while flow_worklist or ssa_worklist:
            while flow_worklist:
                edge = flow_worklist.pop()
                if edge in executable_edges:
                    continue
                executable_edges.add(edge)
                block: BasicBlock = edge.dst
                for p in block.phis:
                    visit_phi(p)
                is_first_visit = sum(edge in executable_edges for edge in block.incoming) <= 1
                if is_first_visit:
                    for stmt in block.statements:
                        if not (isinstance(stmt, IRSet) and isinstance(stmt.place, SSAPlace)):
                            continue
                        value = values[stmt.place]
                        new_value = self.evaluate_stmt(stmt.value, values)
                        if new_value != value:
                            values[stmt.place] = new_value
                            ssa_worklist.update(ssa_edges.get(stmt.place, set()))
                    test_value = values[block]
                    new_test_value = self.evaluate_stmt(block.test, values)
                    if new_test_value != test_value:
                        if new_test_value is UNDEF:
                            continue
                        values[block] = new_test_value
                        if new_test_value is NAC:
                            flow_worklist.update(block.outgoing)
                            reachable_blocks.update(e.dst for e in block.outgoing)
                        elif block.outgoing:
                            outgoing_by_cond = {edge.cond: edge for edge in block.outgoing}
                            if is_constant(new_test_value):
                                taken_edge = outgoing_by_cond.get(new_test_value, outgoing_by_cond.get(None))
                                if taken_edge is None:
                                    raise ValueError("Unexpected missing edge")
                                taken_edges = {taken_edge}
                            else:
                                taken_edges = set()
                                for v in new_test_value:
                                    taken_edge = outgoing_by_cond.get(v, outgoing_by_cond.get(None))
                                    if taken_edge:
                                        taken_edges.add(taken_edge)
                                    else:
                                        raise ValueError("Unexpected missing edge")
                            for taken_edge in taken_edges:
                                flow_worklist.add(taken_edge)
                                reachable_blocks.add(taken_edge.dst)
                    elif len(block.outgoing) == 1 and next(iter(block.outgoing)).cond is None:
                        flow_worklist.update(block.outgoing)
                        reachable_blocks.update(e.dst for e in block.outgoing)
            while ssa_worklist:
                p = ssa_worklist.pop()
                defn = defs[p]
                if isinstance(defn, dict):
                    # This is a phi
                    visit_phi(p)
                elif isinstance(p, BasicBlock):
                    # This is the block's test
                    test_value = values[p]
                    new_test_value = self.evaluate_stmt(defn, values)
                    if new_test_value != test_value:
                        if new_test_value is UNDEF:
                            continue
                        values[p] = new_test_value
                        if new_test_value is NAC:
                            flow_worklist.update(p.outgoing)
                            reachable_blocks.update(e.dst for e in p.outgoing)
                        else:
                            outgoing_by_cond = {edge.cond: edge for edge in p.outgoing}
                            if is_constant(new_test_value):
                                taken_edge = outgoing_by_cond.get(new_test_value, outgoing_by_cond.get(None))
                                if taken_edge is None:
                                    raise ValueError("Unexpected missing edge")
                                taken_edges = {taken_edge}
                            else:
                                taken_edges = set()
                                for v in new_test_value:
                                    taken_edge = outgoing_by_cond.get(v, outgoing_by_cond.get(None))
                                    if taken_edge:
                                        taken_edges.add(taken_edge)
                                    else:
                                        raise ValueError("Unexpected missing edge")
                            for taken_edge in taken_edges:
                                flow_worklist.add(taken_edge)
                                reachable_blocks.add(taken_edge.dst)
                else:
                    # This is a regular SSA assignment
                    if places_to_blocks[p] not in reachable_blocks:
                        continue
                    value = values[p]
                    new_value = self.evaluate_stmt(defn, values)
                    if new_value != value:
                        values[p] = new_value
                        ssa_worklist.update(ssa_edges.get(p, set()))

        for block in traverse_cfg_preorder(entry):
            block.statements = [self.substitute_constants(stmt, values) for stmt in block.statements]
            block.test = self.substitute_constants(block.test, values)
            if isinstance(block.test, IRGet) and block.test.place in values:
                test_value = values[block.test.place]
                if isinstance(test_value, frozenset):
                    new_outgoing = set()
                    outgoing_by_cond = {edge.cond: edge for edge in block.outgoing}
                    for v in test_value:
                        if v in outgoing_by_cond:
                            new_outgoing.add(outgoing_by_cond[v])
                        elif None in outgoing_by_cond:
                            new_outgoing.add(outgoing_by_cond[None])
                        else:
                            raise ValueError("Unexpected missing edge")
                    removed_edges = set(block.outgoing) - new_outgoing
                    for edge in removed_edges:
                        edge.dst.incoming.remove(edge)
                    if not any(edge.cond is None for edge in new_outgoing):
                        default_edge = max(new_outgoing, key=lambda e: e.cond)
                        default_edge.cond = None
                    block.outgoing = new_outgoing

        reachable_blocks = set(traverse_cfg_preorder(entry))
        for block in traverse_cfg_preorder(entry):
            block.incoming = {edge for edge in block.incoming if edge.src in reachable_blocks}
            incoming_blocks = {edge.src for edge in block.incoming}
            for v in block.phis:
                block.phis[v] = {k: v for k, v in block.phis[v].items() if k in incoming_blocks}

        queue = set(traverse_cfg_preorder(entry))
        while queue:
            block = queue.pop()
            dead_phis = {k for k, v in block.phis.items() if not v}
            for edge in block.outgoing:
                for v in edge.dst.phis.values():
                    if block in v and v[block] in dead_phis:
                        del v[block]
                        queue.add(edge.dst)
            block.phis = {k: v for k, v in block.phis.items() if v}

        for block in traverse_cfg_preorder(entry):
            block.phis = {
                k: {b: arg for b, arg in args.items() if values.get(arg, UNDEF) is not UNDEF}
                for k, args in block.phis.items()
            }
            block.statements = [
                stmt
                for stmt in block.statements
                # Note that if this is a set with a side effect, it will never be undef
                if not (
                    isinstance(stmt, IRSet)
                    and isinstance(stmt.place, SSAPlace)
                    and values.get(stmt.place, UNDEF) is UNDEF
                )
            ]

        return entry

    def get_dependencies(self, stmt, dependencies: set[SSAPlace]):
        match stmt:
            case IRConst():
                pass
            case IRPureInstr(op=_, args=args) | IRInstr(op=_, args=args):
                for arg in args:
                    self.get_dependencies(arg, dependencies)
            case IRGet(place=SSAPlace() as place):
                dependencies.add(place)
            case IRGet(place=BlockPlace() as place):
                self.get_dependencies(place.block, dependencies)
            case BlockPlace(block=block, index=index, offset=_):
                self.get_dependencies(block, dependencies)
                self.get_dependencies(index, dependencies)
            case SSAPlace():
                dependencies.add(stmt)
            case int() | float() | TempBlock():
                pass
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")
        return dependencies

    def substitute_constants(self, stmt, values: dict[SSAPlace, Value]):
        match stmt:
            case IRConst():
                return stmt
            case IRPureInstr(op=op, args=args):
                return IRPureInstr(op=op, args=[self.substitute_constants(arg, values) for arg in args])
            case IRInstr(op=op, args=args):
                return IRInstr(op=op, args=[self.substitute_constants(arg, values) for arg in args])
            case IRGet(place=SSAPlace() as place):
                value = values[place]
                if isinstance(value, int | float):
                    return IRConst(value)
                return stmt
            case IRGet(place=place):
                return IRGet(place=self.substitute_constants(place, values))
            case IRSet(place=SSAPlace() as place, value=value):
                return IRSet(place=place, value=self.substitute_constants(value, values))
            case IRSet(place=place, value=value):
                return IRSet(
                    place=self.substitute_constants(place, values), value=self.substitute_constants(value, values)
                )
            case BlockPlace(block=block, index=index, offset=offset):
                return BlockPlace(
                    block=self.substitute_constants(block, values),
                    index=self.substitute_constants(index, values),
                    offset=offset,
                )
            case SSAPlace():
                value = values[stmt]
                if isinstance(value, int | float):
                    return IRConst(value)
                return stmt
            case int() | float() | TempBlock():
                return stmt
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")

    def evaluate_stmt(self, stmt, values: dict[SSAPlace, Value]) -> Value:
        match stmt:
            case IRConst(value=value):
                return value
            case IRPureInstr(op=op, args=args) | IRInstr(op=op, args=args):
                if op not in self.SUPPORTED_OPS:
                    return NAC
                args = [self.evaluate_stmt(arg, values) for arg in args]
                match op:
                    case Op.And:
                        if any(arg == 0 for arg in args):
                            return 0
                    case Op.Or:
                        if any(arg == 1 for arg in args):
                            return 1
                    case Op.Multiply:
                        if any(arg == 0 for arg in args):
                            return 0
                if any(arg is NAC or isinstance(arg, frozenset) for arg in args):
                    return NAC
                if any(arg is UNDEF for arg in args):
                    return UNDEF
                match op:
                    case Op.Equal:
                        assert len(args) == 2
                        return args[0] == args[1]
                    case Op.NotEqual:
                        assert len(args) == 2
                        return args[0] != args[1]
                    case Op.Greater:
                        assert len(args) == 2
                        return args[0] > args[1]
                    case Op.GreaterOr:
                        assert len(args) == 2
                        return args[0] >= args[1]
                    case Op.Less:
                        assert len(args) == 2
                        return args[0] < args[1]
                    case Op.LessOr:
                        assert len(args) == 2
                        return args[0] <= args[1]
                    case Op.Not:
                        assert len(args) == 1
                        return int(not args[0])
                    case Op.And:
                        return all(args)
                    case Op.Or:
                        return any(args)
                    case Op.Negate:
                        assert len(args) == 1
                        return -args[0]
                    case Op.Add:
                        return sum(args)
                    case Op.Subtract:
                        if len(args) == 0:
                            return 0
                        return args[0] - sum(args[1:])
                    case Op.Multiply:
                        if len(args) == 0:
                            return 1
                        return functools.reduce(operator.mul, args, 1)
                    case Op.Divide:
                        if len(args) == 0:
                            return 1
                        return args[0] / functools.reduce(operator.mul, args[1:], 1)
                    case Op.Power:
                        if len(args) == 0:
                            return 1
                        return functools.reduce(operator.pow, args)
                    case Op.Log:
                        assert len(args) == 1
                        return math.log(args[0])
                    case Op.Ceil:
                        assert len(args) == 1
                        return math.ceil(args[0])
                    case Op.Floor:
                        assert len(args) == 1
                        return math.floor(args[0])
                    case Op.Round:
                        assert len(args) == 1
                        # This is round half to even in both Python and Sonolus
                        return round(args[0])
                    case Op.Frac:
                        assert len(args) == 1
                        return smath.frac(args[0])
                    case Op.Mod:
                        assert len(args) == 2
                        return args[0] % args[1]
                    case Op.Rem:
                        assert len(args) == 2
                        return smath.remainder(args[0], args[1])
                    case Op.Sin:
                        assert len(args) == 1
                        return math.sin(args[0])
                    case Op.Cos:
                        assert len(args) == 1
                        return math.cos(args[0])
                    case Op.Tan:
                        assert len(args) == 1
                        return math.tan(args[0])
                    case Op.Sinh:
                        assert len(args) == 1
                        return math.sinh(args[0])
                    case Op.Cosh:
                        assert len(args) == 1
                        return math.cosh(args[0])
                    case Op.Tanh:
                        assert len(args) == 1
                        return math.tanh(args[0])
                    case Op.Arcsin:
                        assert len(args) == 1
                        return math.asin(args[0])
                    case Op.Arccos:
                        assert len(args) == 1
                        return math.acos(args[0])
                    case Op.Arctan:
                        assert len(args) == 1
                        return math.atan(args[0])
                    case Op.Arctan2:
                        assert len(args) == 2
                        return math.atan2(args[0], args[1])
                    case Op.Max:
                        assert len(args) == 2
                        return max(args)
                    case Op.Min:
                        assert len(args) == 2
                        return min(args)
                    case Op.Abs:
                        assert len(args) == 1
                        return abs(args[0])
                    case Op.Clamp:
                        assert len(args) == 3
                        return min(max(args[0], args[1]), args[2])
                    case Op.Degree:
                        assert len(args) == 1
                        return math.degrees(args[0])
                    case Op.Radian:
                        assert len(args) == 1
                        return math.radians(args[0])
                    case Op.Lerp:
                        assert len(args) == 3
                        return args[0] + (args[1] - args[0]) * args[2]
                    case Op.LerpClamped:
                        assert len(args) == 3
                        return args[0] + (args[1] - args[0]) * max(0, min(1, args[2]))
                    case Op.Remap:
                        assert len(args) == 5
                        try:
                            return args[2] + (args[3] - args[2]) * (args[4] - args[0]) / (args[1] - args[0])
                        except (ValueError, ZeroDivisionError, OverflowError):
                            return NAC
                    case Op.RemapClamped:
                        assert len(args) == 5
                        try:
                            return args[2] + (args[3] - args[2]) * max(
                                0, min(1, (args[4] - args[0]) / (args[1] - args[0]))
                            )
                        except (ValueError, ZeroDivisionError, OverflowError):
                            return NAC
            case IRGet(place=SSAPlace() as place):
                return values[place]
            case IRGet():
                return NAC
            case IRSet() | _:
                raise TypeError(f"Unexpected statement: {stmt}")



from math import floor, pi, trunc
from typing import Literal

from pydori.lib.layout import END_LANE, LANE_COUNT, START_LANE
from sonolus.script.globals import level_data
from sonolus.script.interval import lerp
from sonolus.script.printing import PrintColor, PrintFormat, print_number
from sonolus.script.quad import Quad, Rect
from sonolus.script.runtime import HorizontalAlign, ScrollDirection, canvas, screen
from sonolus.script.vec import Vec2

# Number of seconds each column represents.
PREVIEW_COLUMN_SECS = 2

# Gap to add at top and bottom of the screen.
PREVIEW_MARGIN_Y = 0.1

# Gap to add on each side of each column (i.e. double this for the total gap between columns).
PREVIEW_MARGIN_X = 0.25

# Gap to add between text and lanes and edges of each column.
PREVIEW_TEXT_MARGIN_X = 0.015

# Gap to add between text and bar line positions.
PREVIEW_TEXT_MARGIN_Y = 0

# Height of text to be printed.
PREVIEW_TEXT_HEIGHT = 0.12

# Width of text to be printed.
PREVIEW_TEXT_WIDTH = PREVIEW_MARGIN_X - 2 * PREVIEW_TEXT_MARGIN_X

# Width of each lane.
PREVIEW_LANE_WIDTH = 0.072

# Width of stage borders.
PREVIEW_STAGE_BORDER_WIDTH = 0.25 * PREVIEW_LANE_WIDTH

# Width of each note.
PREVIEW_NOTE_WIDTH = PREVIEW_LANE_WIDTH

# Height of bar lines.
PREVIEW_BAR_LINE_HEIGHT = 0.0036

# Alpha of bar lines.
PREVIEW_BAR_LINE_ALPHA = 0.8

# Alpha of the top and bottom cover.
PREVIEW_COVER_ALPHA = 1.0

# The y-coordinate of the bottom of the screen plus the margin.
PREVIEW_Y_MIN = -1 + PREVIEW_MARGIN_Y

# The y-coordinate of the top of the screen minus the margin.
PREVIEW_Y_MAX = 1 - PREVIEW_MARGIN_Y

# Amount bar lines are extended on each side if configured to do so.
PREVIEW_BAR_EXTEND_WIDTH = 3 * PREVIEW_LANE_WIDTH

# Offset added to flick arrows to make them appear above their notes.
PREVIEW_FLICK_ARROW_Y_OFFSET = 0.9 * PREVIEW_NOTE_WIDTH


@level_data
class PreviewData:
    highest_lane: float
    last_time: float
    last_beat: float


@level_data
class PreviewLayout:
    column_count: int
    column_width: float


def init_preview_layout():
    PreviewLayout.column_count = floor(PreviewData.last_time / PREVIEW_COLUMN_SECS) + 1
    PreviewLayout.column_width = 2 * PREVIEW_MARGIN_X + PREVIEW_LANE_WIDTH * LANE_COUNT

    canvas().update(
        scroll_direction=ScrollDirection.LEFT_TO_RIGHT,
        size=PreviewLayout.column_width * PreviewLayout.column_count,
    )


def time_to_preview_col(time: float) -> int:
    return trunc(time / PREVIEW_COLUMN_SECS)


def time_to_preview_y(time: float) -> float:
    return lerp(PREVIEW_Y_MIN, PREVIEW_Y_MAX, time % PREVIEW_COLUMN_SECS / PREVIEW_COLUMN_SECS)


def lane_to_preview_x(lane: float, col: int) -> float:
    return (col + 0.5) * PreviewLayout.column_width + lane * PREVIEW_LANE_WIDTH - screen().w / 2


def lane_to_preview_left_x(lane: float, col: int) -> float:
    return lane_to_preview_x(lane - 0.5, col)


def lane_to_preview_right_x(lane: float, col: int) -> float:
    return lane_to_preview_x(lane + 0.5, col)


def layout_preview_lane(lane: float, col: int) -> Rect:
    return Rect(
        l=lane_to_preview_left_x(lane, col),
        r=lane_to_preview_right_x(lane, col),
        t=1,
        b=-1,
    )


def layout_preview_stage_border_left(col: int) -> Rect:
    return Rect(
        l=lane_to_preview_left_x(START_LANE, col) - PREVIEW_STAGE_BORDER_WIDTH,
        r=lane_to_preview_left_x(START_LANE, col),
        t=1,
        b=-1,
    )


def layout_preview_stage_border_right(col: int) -> Rect:
    return Rect(
        l=lane_to_preview_right_x(END_LANE, col),
        r=lane_to_preview_right_x(END_LANE, col) + PREVIEW_STAGE_BORDER_WIDTH,
        t=1,
        b=-1,
    )


def layout_preview_note(lane: float, time: float) -> Rect:
    col = time_to_preview_col(time)
    y = time_to_preview_y(time)
    return Rect(
        l=lane_to_preview_left_x(lane, col),
        r=lane_to_preview_right_x(lane, col),
        t=y + PREVIEW_NOTE_WIDTH / 2,
        b=y - PREVIEW_NOTE_WIDTH / 2,
    )


def layout_preview_flick_arrow(lane: float, time: float) -> Quad:
    col = time_to_preview_col(time)
    left_x = lane_to_preview_left_x(lane, col)
    right_x = lane_to_preview_right_x(lane, col)
    y = time_to_preview_y(time) + PREVIEW_FLICK_ARROW_Y_OFFSET
    return Quad(
        bl=Vec2(left_x, y - PREVIEW_NOTE_WIDTH / 2),
        br=Vec2(right_x, y - PREVIEW_NOTE_WIDTH / 2),
        tr=Vec2(right_x, y + PREVIEW_NOTE_WIDTH / 2),
        tl=Vec2(left_x, y + PREVIEW_NOTE_WIDTH / 2),
    )


def layout_preview_directional_flick_arrow(lane: float, time: float, direction: float) -> Quad:
    col = time_to_preview_col(time)
    left_x = lane_to_preview_left_x(lane, col)
    right_x = lane_to_preview_right_x(lane, col)
    y = time_to_preview_y(time)
    return (
        Rect(
            l=left_x,
            r=right_x,
            t=y + PREVIEW_NOTE_WIDTH / 2,
            b=y - PREVIEW_NOTE_WIDTH / 2,
        )
        .as_quad()
        .rotate_centered(
            -pi / 2 if direction > 0 else pi / 2,
        )
    )


def layout_preview_connector(
    lane_a: float,
    lane_b: float,
    time_a: float,
    time_b: float,
    col: int,
) -> Quad:
    col_time = col * PREVIEW_COLUMN_SECS
    left_x_a = lane_to_preview_left_x(lane_a, col)
    right_x_a = lane_to_preview_right_x(lane_a, col)
    left_x_b = lane_to_preview_left_x(lane_b, col)
    right_x_b = lane_to_preview_right_x(lane_b, col)
    y_a = lerp(
        PREVIEW_Y_MIN,
        PREVIEW_Y_MAX,
        (time_a - col_time) / PREVIEW_COLUMN_SECS,
    )
    y_b = lerp(
        PREVIEW_Y_MIN,
        PREVIEW_Y_MAX,
        (time_b - col_time) / PREVIEW_COLUMN_SECS,
    )
    return Quad(
        bl=Vec2(left_x_a, y_a),
        tl=Vec2(left_x_b, y_b),
        tr=Vec2(right_x_b, y_b),
        br=Vec2(right_x_a, y_a),
    )


def layout_preview_sim_line(
    lane_a: float,
    lane_b: float,
    time: float,
) -> Quad:
    col = time_to_preview_col(time)
    left_x = lane_to_preview_x(lane_a, col)
    right_x = lane_to_preview_x(lane_b, col)
    y = time_to_preview_y(time)
    return Quad(
        bl=Vec2(left_x, y - PREVIEW_LANE_WIDTH / 2 / 4),
        br=Vec2(right_x, y - PREVIEW_LANE_WIDTH / 2 / 4),
        tr=Vec2(right_x, y + PREVIEW_LANE_WIDTH / 2 / 4),
        tl=Vec2(left_x, y + PREVIEW_LANE_WIDTH / 2 / 4),
    )


def layout_preview_bar_line(
    time: float,
    extend: Literal["left", "right", "both", "none", "left_only", "right_only"] = "none",
) -> Quad:
    col = time_to_preview_col(time)
    left_lane = START_LANE - 0.5
    right_lane = END_LANE + 0.5
    left_x = lane_to_preview_x(left_lane, col)
    right_x = lane_to_preview_x(right_lane, col)
    match extend:
        case "left":
            left_x -= PREVIEW_BAR_EXTEND_WIDTH
        case "right":
            right_x += PREVIEW_BAR_EXTEND_WIDTH
        case "both":
            left_x -= PREVIEW_BAR_EXTEND_WIDTH
            right_x += PREVIEW_BAR_EXTEND_WIDTH
        case "left_only":
            right_x = left_x
            left_x -= PREVIEW_BAR_EXTEND_WIDTH
        case "right_only":
            left_x = right_x
            right_x += PREVIEW_BAR_EXTEND_WIDTH
        case _:
            pass
    y = time_to_preview_y(time)
    return Quad(
        bl=Vec2(left_x, y - PREVIEW_BAR_LINE_HEIGHT / 2),
        br=Vec2(right_x, y - PREVIEW_BAR_LINE_HEIGHT / 2),
        tr=Vec2(right_x, y + PREVIEW_BAR_LINE_HEIGHT / 2),
        tl=Vec2(left_x, y + PREVIEW_BAR_LINE_HEIGHT / 2),
    )


def print_at_time(
    value: float,
    time: float,
    *,
    fmt: PrintFormat,
    decimal_places: int = -1,
    color: PrintColor,
    side: Literal["left", "right"],
):
    print_number(
        value=value,
        fmt=fmt,
        decimal_places=decimal_places,
        anchor=Vec2(
            lane_to_preview_x(
                -LANE_COUNT / 2 if side == "left" else LANE_COUNT / 2,
                time_to_preview_col(time),
            )
            + (-PREVIEW_TEXT_MARGIN_X if side == "left" else PREVIEW_TEXT_MARGIN_X),
            time_to_preview_y(time) + PREVIEW_TEXT_MARGIN_Y,
        ),
        pivot=Vec2(1 if side == "left" else 0, 0),
        dimensions=Vec2(PREVIEW_TEXT_WIDTH, PREVIEW_TEXT_HEIGHT),
        color=color,
        horizontal_align=HorizontalAlign.RIGHT if side == "left" else HorizontalAlign.LEFT,
        background=False,
    )



from sonolus.backend.ir import IRGet
from sonolus.backend.place import BlockPlace
from sonolus.script.internal.context import ctx
from sonolus.script.internal.generic import validate_concrete_type
from sonolus.script.internal.impl import validate_value
from sonolus.script.internal.meta_fn import meta_fn
from sonolus.script.internal.value import BackingValue, ExprBackingValue, ReadOnlyBackingValueWrapper


@meta_fn
def alloc[T](type_: type[T]) -> T:  # type: ignore
    """Return an uninitialized instance of the given type.

    Use this carefully as reading from uninitialized memory can lead to unexpected behavior.
    """
    type_ = validate_concrete_type(type_)
    if ctx():
        return type_._alloc_()  # type: ignore
    else:
        return type_._alloc_()._as_py_()  # type: ignore


@meta_fn
def zeros[T](type_: type[T]) -> T:  # type: ignore
    """Make a new instance of the given type initialized with zeros.

    Generally works the same as the unary `+` operator on record and array types.
    """
    return validate_concrete_type(type_)._zero_()  # type: ignore


@meta_fn
def copy[T](value: T) -> T:
    """Make a deep copy of the given value.

    Generally works the same as the unary `+` operator on records and arrays.
    """
    value = validate_value(value)  # type: ignore
    if ctx():
        return value._copy_()  # type: ignore
    else:
        return value._copy_()._as_py_()  # type: ignore


def swap[T](a: T, b: T):
    """Swap the values of the two provided mutable values."""
    temp = copy(a)
    a @= b  # type: ignore
    b @= temp  # type: ignore


@meta_fn
def sizeof(type_: type, /) -> int:
    """Return the size of the given type."""
    return validate_concrete_type(type_)._size_()


@meta_fn
def freeze[T](value: T) -> T:
    """Return a read-only version of the given value.

    If the original value is modified, no guarantees are made about behavior and changes may or may not be reflected
    in the frozen version. Use freeze(copy(value)) to guarantee independence.
    """
    value = validate_value(value)
    result = type(value)._from_list_(_map_data_value_to_frozen(v) for v in value._to_list_())
    if ctx():
        return result
    else:
        return result._as_py_()


def _map_data_value_to_frozen(v):
    match v:
        case int() | float() | bool() | ExprBackingValue() | ReadOnlyBackingValueWrapper():
            return v
        case BackingValue():
            return ReadOnlyBackingValueWrapper(v)
        case BlockPlace():
            return ExprBackingValue(IRGet(v))
        case _:
            raise TypeError("Unsupported value for freezing")
